@@ -1,4 +1,4 @@
-/* Copyright (c) 2000, 2011, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2000, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -10,8 +10,8 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
+   along with this program; if not, write to the Free Software Foundation,
+   51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA */
 
 
 /**
@@ -82,11 +82,10 @@
 #include "sql_acl.h"                       // SUPER_ACL
 #include <hash.h>
 #include <assert.h>
-
+#include "my_atomic.h"
 #ifdef WITH_WSREP
 #include "wsrep_mysqld.h"
 #endif /* WITH_WSREP */
-
 /**
   @defgroup Locking Locking
   @{
@@ -119,15 +118,20 @@ static int thr_lock_errno_to_mysql[]=
 static int
 lock_tables_check(THD *thd, TABLE **tables, uint count, uint flags)
 {
-  uint system_count, i;
-  bool is_superuser, log_table_write_query;
+  uint system_count= 0, i= 0;
+  bool enforce_ro= true;
+  /*
+    Identifies if the executed sql command can updated either a log
+    or rpl info table.
+  */
+  bool log_table_write_query= false;
 
   DBUG_ENTER("lock_tables_check");
 
-  system_count= 0;
-  is_superuser= thd->security_ctx->master_access & SUPER_ACL;
-  log_table_write_query= (is_log_table_write_query(thd->lex->sql_command)
-                         || ((flags & MYSQL_LOCK_LOG_TABLE) != 0));
+  if (!opt_super_readonly)
+    enforce_ro= !(thd->security_ctx->master_access & SUPER_ACL);
+  log_table_write_query=
+     is_log_table_write_query(thd->lex->sql_command);
 
   for (i=0 ; i<count; i++)
   {
@@ -142,16 +146,17 @@ lock_tables_check(THD *thd, TABLE **tables, uint count, uint flags)
       When a user is requesting a lock, the following
       constraints are enforced:
     */
-    if (t->s->require_write_privileges() &&
-        ! log_table_write_query)
+    if (t->s->table_category == TABLE_CATEGORY_LOG &&
+        (flags & MYSQL_LOCK_LOG_TABLE) == 0 &&
+        !log_table_write_query)
     {
       /*
         A user should not be able to prevent writes,
         or hold any type of lock in a session,
         since this would be a DOS attack.
       */
-      if ((t->reginfo.lock_type >= TL_READ_NO_INSERT)
-          || (thd->lex->sql_command == SQLCOM_LOCK_TABLES))
+      if (t->reginfo.lock_type >= TL_READ_NO_INSERT ||
+          thd->lex->sql_command == SQLCOM_LOCK_TABLES)
       {
         my_error(ER_CANT_LOCK_LOG_TABLE, MYF(0));
         DBUG_RETURN(1);
@@ -196,9 +201,14 @@ lock_tables_check(THD *thd, TABLE **tables, uint count, uint flags)
     if (!(flags & MYSQL_LOCK_IGNORE_GLOBAL_READ_ONLY) && !t->s->tmp_table)
     {
       if (t->reginfo.lock_type >= TL_WRITE_ALLOW_WRITE &&
-          !is_superuser && opt_readonly && !thd->slave_thread)
+          enforce_ro && opt_readonly && !thd->slave_thread
+#ifdef WITH_WSREP
+          && !thd->wsrep_applier
+#endif /* WITH_WSREP */
+         )
       {
-        my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0), "--read-only");
+        my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0),
+                 opt_super_readonly ? "--read-only (super)" : "--read-only");
         DBUG_RETURN(1);
       }
     }
@@ -300,7 +310,7 @@ MYSQL_LOCK *mysql_lock_tables(THD *thd, TABLE **tables, uint count, uint flags)
   if (! (sql_lock= get_lock_data(thd, tables, count, GET_LOCK_STORE_LOCKS)))
     DBUG_RETURN(NULL);
 
-  thd_proc_info(thd, "System lock");
+  THD_STAGE_INFO(thd, stage_system_lock);
   DBUG_PRINT("info", ("thd->proc_info %s", thd->proc_info));
   if (sql_lock->table_count && lock_external(thd, sql_lock->table,
                                              sql_lock->table_count))
@@ -331,13 +341,7 @@ MYSQL_LOCK *mysql_lock_tables(THD *thd, TABLE **tables, uint count, uint flags)
       my_error(rc, MYF(0));
   }
 end:
-#ifdef WITH_WSREP
-  thd_proc_info(thd, "mysql_lock_tables(): unlocking tables II");
-#else /* WITH_WSREP */
-  thd_proc_info(thd, 0);
-#endif /* WITH_WSREP */
-
-  if (thd->killed)
+  if (!(flags & MYSQL_OPEN_IGNORE_KILLED) && thd->killed)
   {
     thd->send_kill_message();
     if (sql_lock)
@@ -348,9 +352,6 @@ end:
   }
 
   thd->set_time_after_lock();
-#ifdef WITH_WSREP
-  thd_proc_info(thd, "exit mysqld_lock_tables()");
-#endif /* WITH_WSREP */
   DBUG_RETURN(sql_lock);
 }
 
@@ -738,8 +739,13 @@ static MYSQL_LOCK *get_lock_data(THD *thd, TABLE **table_ptr, uint count,
     }
     *to++= table;
     if (locks)
+    {
       for ( ; org_locks != locks ; org_locks++)
+      {
 	(*org_locks)->debug_print_param= (void *) table;
+        (*org_locks)->m_psi= table->file->m_psi;
+      }
+    }
   }
   /*
     We do not use 'tables', because there are cases where store_lock()
@@ -783,6 +789,7 @@ bool lock_schema_name(THD *thd, const char *db)
 {
   MDL_request_list mdl_requests;
   MDL_request global_request;
+  MDL_request backup_request;
   MDL_request mdl_request;
 
   if (thd->locked_tables_mode)
@@ -796,9 +803,16 @@ bool lock_schema_name(THD *thd, const char *db)
     return TRUE;
   global_request.init(MDL_key::GLOBAL, "", "", MDL_INTENTION_EXCLUSIVE,
                       MDL_STATEMENT);
+
+  if (thd->backup_tables_lock.abort_if_acquired())
+    return true;
+  thd->backup_tables_lock.init_protection_request(&backup_request,
+                                                  MDL_STATEMENT);
+
   mdl_request.init(MDL_key::SCHEMA, db, "", MDL_EXCLUSIVE, MDL_TRANSACTION);
 
   mdl_requests.push_front(&mdl_request);
+  mdl_requests.push_front(&backup_request);
   mdl_requests.push_front(&global_request);
 
   if (thd->mdl_context.acquire_locks(&mdl_requests,
@@ -874,35 +888,35 @@ bool lock_object_name(THD *thd, MDL_key::enum_mdl_namespace mdl_type,
 
 static void print_lock_error(int error, const char *table)
 {
-  int textno;
   DBUG_ENTER("print_lock_error");
 
   switch (error) {
   case HA_ERR_LOCK_WAIT_TIMEOUT:
-    textno=ER_LOCK_WAIT_TIMEOUT;
+    my_error(ER_LOCK_WAIT_TIMEOUT, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG), error);
     break;
   case HA_ERR_READ_ONLY_TRANSACTION:
-    textno=ER_READ_ONLY_TRANSACTION;
+    my_error(ER_READ_ONLY_TRANSACTION, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG),
+             error);
     break;
   case HA_ERR_LOCK_DEADLOCK:
-    textno=ER_LOCK_DEADLOCK;
+    my_error(ER_LOCK_DEADLOCK, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG), error);
     break;
   case HA_ERR_WRONG_COMMAND:
-    textno=ER_ILLEGAL_HA;
+    my_error(ER_ILLEGAL_HA, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG), table);
     break;
   default:
-    textno=ER_CANT_LOCK;
+    {
+      char errbuf[MYSYS_STRERROR_SIZE];
+      my_error(ER_CANT_LOCK, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG),
+               error, my_strerror(errbuf, sizeof(errbuf), error));
+    }
     break;
   }
 
-  if ( textno == ER_ILLEGAL_HA )
-    my_error(textno, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG), table);
-  else
-    my_error(textno, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG), error);
-
   DBUG_VOID_RETURN;
 }
 
+volatile int32 Global_read_lock::m_active_requests;
 
 /****************************************************************************
   Handling of global read locks
@@ -971,24 +985,44 @@ static void print_lock_error(int error, const char *table)
   @retval True   Failure, thread was killed.
 */
 
-bool Global_read_lock::lock_global_read_lock(THD *thd)
+bool Global_read_lock::lock_global_read_lock(THD *thd, bool *own_lock)
 {
   DBUG_ENTER("lock_global_read_lock");
 
+  *own_lock= FALSE;
+
   if (!m_state)
   {
     MDL_request mdl_request;
 
     DBUG_ASSERT(! thd->mdl_context.is_lock_owner(MDL_key::GLOBAL, "", "",
                                                  MDL_SHARED));
+
+    /*
+       Do not allow upgrading backup locks to FTWRL. Otherwise we can end up
+       with deadlocks being reported for queries previously blocked on a backup
+       lock.
+    */
+    if (thd->backup_tables_lock.abort_if_acquired() ||
+        thd->backup_binlog_lock.abort_if_acquired())
+      DBUG_RETURN(1);
+
     mdl_request.init(MDL_key::GLOBAL, "", "", MDL_SHARED, MDL_EXPLICIT);
 
+    /* Increment static variable first to signal innodb memcached server
+       to release mdl locks held by it */
+    my_atomic_add32(&Global_read_lock::m_active_requests, 1);
     if (thd->mdl_context.acquire_lock(&mdl_request,
                                       thd->variables.lock_wait_timeout))
+    {
+      my_atomic_add32(&Global_read_lock::m_active_requests, -1);
       DBUG_RETURN(1);
+    }
 
     m_mdl_global_shared_lock= mdl_request.ticket;
     m_state= GRL_ACQUIRED;
+
+    *own_lock= TRUE;
   }
   /*
     We DON'T set global_read_lock_blocks_commit now, it will be set after
@@ -1023,14 +1057,35 @@ void Global_read_lock::unlock_global_read_lock(THD *thd)
     thd->mdl_context.release_lock(m_mdl_blocks_commits_lock);
     m_mdl_blocks_commits_lock= NULL;
 #ifdef WITH_WSREP
-    wsrep_locked_seqno= WSREP_SEQNO_UNDEFINED;
-    wsrep->resume(wsrep);
+    if (thd->wsrep_sst_donor)
+    {
+      /* If this is sst_donor then it is resync internally.
+      So only resume the cluster. */
+      wsrep_resume();
+    }
+    else if (WSREP(thd) && provider_desynced_paused)
+    {
+      /* Function will take care of decrementing reference count
+      if it is not the last one to get called. Last one will
+      perform the resume action. */
+      wsrep_resume();
+
+      int ret = wsrep->resync(wsrep);
+      if (ret != WSREP_OK)
+      {
+        WSREP_WARN("resync failed %d for FTWRL: db: %s, query: %s", ret,
+                   (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
+        DBUG_VOID_RETURN;
+      }
+      provider_desynced_paused= false;
+    }
 #endif /* WITH_WSREP */
   }
+
   thd->mdl_context.release_lock(m_mdl_global_shared_lock);
+  my_atomic_add32(&Global_read_lock::m_active_requests, -1);
   m_mdl_global_shared_lock= NULL;
   m_state= GRL_NONE;
- 
   DBUG_VOID_RETURN;
 }
 
@@ -1058,22 +1113,19 @@ bool Global_read_lock::make_global_read_lock_block_commit(THD *thd)
     If we didn't succeed lock_global_read_lock(), or if we already suceeded
     make_global_read_lock_block_commit(), do nothing.
   */
-  if (m_state != GRL_ACQUIRED)
-    DBUG_RETURN(0);
 
 #ifdef WITH_WSREP
-  if (m_mdl_blocks_commits_lock)
+  if (WSREP(thd) && m_mdl_blocks_commits_lock)
   {
     WSREP_DEBUG("GRL was in block commit mode when entering "
 		"make_global_read_lock_block_commit");
-    thd->mdl_context.release_lock(m_mdl_blocks_commits_lock);
-    m_mdl_blocks_commits_lock= NULL;
-    wsrep_locked_seqno= WSREP_SEQNO_UNDEFINED;
-    wsrep->resume(wsrep);
-    m_state= GRL_ACQUIRED;
+    DBUG_RETURN(FALSE);
   }
 #endif /* WITH_WSREP */
 
+  if (m_state != GRL_ACQUIRED)
+    DBUG_RETURN(0);
+
   mdl_request.init(MDL_key::COMMIT, "", "", MDL_SHARED, MDL_EXPLICIT);
 
   if (thd->mdl_context.acquire_lock(&mdl_request,
@@ -1084,24 +1136,114 @@ bool Global_read_lock::make_global_read_lock_block_commit(THD *thd)
   m_state= GRL_ACQUIRED_AND_BLOCKS_COMMIT;
 
 #ifdef WITH_WSREP
-  long long ret = wsrep->pause(wsrep);
+  if (thd->wsrep_sst_donor)
+  {
+    /* If this is sst_donor then it is already desync internally.
+    So only pause the cluster */
+    if (!wsrep_pause())
+      DBUG_RETURN(TRUE);
+  }
+  else if (WSREP(thd) && !provider_desynced_paused)
+  {
+    int rcode;
+    WSREP_DEBUG("running implicit desync for node");
+    rcode = wsrep->desync(wsrep);
+    if (rcode != WSREP_OK)
+    {
+      WSREP_WARN("FTWRL desync failed %d for schema: %s, query: %s",
+                 rcode, (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
+      my_message(ER_LOCK_DEADLOCK, "wsrep desync failed for FTWRL", MYF(0));
+      DBUG_RETURN(TRUE);
+    }
+
+    if (!wsrep_pause())
+    {
+      /* pause failed so rollback desync action too. */
+      wsrep->resync(wsrep);
+      DBUG_RETURN(TRUE);
+    }
+
+    provider_desynced_paused= true;
+  }
+#endif /* WITH_WSREP */
+
+  DBUG_RETURN(FALSE);
+}
+
+#ifdef WITH_WSREP
+/**
+  Pause the galera provider.
+  Also set wsrep_locked_seqno to sequence number returned.
+
+  @retval False  Failed to pause the provider, wsrep_locked_seqno is reset.
+  @retval True   Provider has been paused.
+*/
+bool Global_read_lock::wsrep_pause()
+{
+  wsrep_seqno_t ret = wsrep->pause(wsrep);
   if (ret >= 0)
   {
     wsrep_locked_seqno= ret;
+    pause_provider(true);
   }
   else if (ret != -ENOSYS) /* -ENOSYS - no provider */
   {
-    WSREP_ERROR("Failed to pause provider: %lld (%s)", -ret, strerror(-ret));
+    WSREP_ERROR("Failed to pause provider: %lld (%s)", (long long)-ret, strerror(-ret));
 
     /* m_mdl_blocks_commits_lock is always NULL here */
     wsrep_locked_seqno= WSREP_SEQNO_UNDEFINED;
     my_error(ER_LOCK_DEADLOCK, MYF(0));
-    DBUG_RETURN(TRUE);
+    return FALSE;
   }
-#endif /* WITH_WSREP */
-  DBUG_RETURN(FALSE);
+
+  return TRUE;
 }
 
+/**
+  Resume the galera provider.
+  Also set wsrep_locked_seqno to sequence number returned.
+
+  @retval False  Failed to pause the provider, wsrep_locked_seqno is reset.
+  @retval True   Provider has been paused.
+*/
+wsrep_status_t Global_read_lock::wsrep_resume()
+{
+  wsrep_status_t ret;
+
+  wsrep_locked_seqno= WSREP_SEQNO_UNDEFINED;
+  ret = wsrep->resume(wsrep);
+
+  if (ret != WSREP_OK)
+  {
+    WSREP_WARN("Failed to resume provider: %d", ret);
+  }
+  else
+  {
+    pause_provider(false);
+  }
+
+  return ret;
+}
+
+bool Global_read_lock::wsrep_pause_once(bool *already_paused)
+{
+    if (!provider_paused)
+    {
+      *already_paused= FALSE;
+      return wsrep_pause();
+    }
+    *already_paused= TRUE;
+    return TRUE;
+}
+
+wsrep_status_t Global_read_lock::wsrep_resume_once(void)
+{
+    if (provider_paused)
+      return wsrep_resume();
+    return WSREP_OK;
+}
+
+#endif /* WITH_WSREP */
 
 /**
   Set explicit duration for metadata locks which are used to implement GRL.
@@ -1117,6 +1259,160 @@ void Global_read_lock::set_explicit_lock_duration(THD *thd)
     thd->mdl_context.set_lock_duration(m_mdl_blocks_commits_lock, MDL_EXPLICIT);
 }
 
+/**
+   Acquire a global backup lock. Wait until all protection requests are
+   released.
+
+  @param thd     Reference to connection.
+
+  @retval false  Success, global backup lock set.
+  @retval true   Failure, thread was killed or lock wait timed out.
+*/
+
+bool Global_backup_lock::acquire(THD *thd)
+{
+  MDL_request mdl_request;
+
+  DBUG_ENTER("Global_backup_lock::acquire");
+
+  DBUG_ASSERT(m_lock == NULL &&
+              !thd->mdl_context.is_lock_owner(m_namespace, "", "",
+                                              MDL_SHARED));
+
+  mdl_request.init(m_namespace, "", "", MDL_SHARED, MDL_EXPLICIT);
+
+  if (thd->mdl_context.acquire_lock(&mdl_request,
+                                    thd->variables.lock_wait_timeout))
+    DBUG_RETURN(true);
+
+  m_lock= mdl_request.ticket;
+
+  DBUG_RETURN(false);
+}
+
+/**
+   Release a global backup lock.
+
+   @param thd     Reference to connection.
+*/
+
+void Global_backup_lock::release(THD *thd)
+{
+  DBUG_ENTER("Global_backup_lock::release");
+
+  DBUG_ASSERT(m_lock != NULL &&
+              thd->mdl_context.is_lock_owner(m_namespace, "", "",
+                                             MDL_SHARED));
+
+  thd->mdl_context.release_lock(m_lock);
+
+  m_lock= NULL;
+
+  DBUG_VOID_RETURN;
+}
+
+/**
+   Set explicit lock duration for the backup metadata lock.
+
+   @param thd     Reference to connection.
+*/
+
+void Global_backup_lock::set_explicit_locks_duration(THD *thd)
+{
+  DBUG_ENTER("Global_backup_lock::set_explicit_lock_duration");
+
+  if (m_lock)
+    thd->mdl_context.set_lock_duration(m_lock, MDL_EXPLICIT);
+
+  DBUG_ASSERT((m_lock != NULL) ==
+              thd->mdl_context.is_lock_owner(m_namespace, "", "",
+                                             MDL_SHARED));
+
+  if (m_prot_lock)
+    thd->mdl_context.set_lock_duration(m_prot_lock, MDL_EXPLICIT);
+
+  DBUG_ASSERT((m_prot_lock != NULL) ==
+              thd->mdl_context.is_lock_owner(m_namespace, "", "",
+                                             MDL_INTENTION_EXCLUSIVE));
+
+  DBUG_VOID_RETURN;
+}
+
+/**
+   Acquire protection against a global backup lock. Wait if a global backup lock
+   is active.
+
+   @param thd               Reference to connection.
+   @param duration          MDL lock duration
+   @param lock_wait_timeout Lock wait timeout
+
+   @retval false  Success, protection was acquired.
+   @retval true   Failure, thread was killed or lock wait timed out.
+*/
+
+bool Global_backup_lock::acquire_protection(THD *thd,
+                                            enum_mdl_duration duration,
+                                            ulong lock_wait_timeout)
+{
+  MDL_request mdl_request;
+
+  DBUG_ENTER("Global_backup_lock::acquire_protection");
+
+  DBUG_ASSERT(duration != MDL_EXPLICIT ||
+              !thd->mdl_context.is_lock_owner(m_namespace, "", "",
+                                              MDL_INTENTION_EXCLUSIVE));
+
+  init_protection_request(&mdl_request, duration);
+
+  if (thd->mdl_context.acquire_lock(&mdl_request, lock_wait_timeout))
+    DBUG_RETURN(true);
+
+  if (duration == MDL_EXPLICIT)
+    m_prot_lock= mdl_request.ticket;
+
+  DBUG_RETURN(false);
+}
+
+/**
+   Initialize a protection request to be acquired later.
+
+   @param mdl_request       Pointer to request to initialize
+   @param duration          MDL lock duration
+*/
+
+void
+Global_backup_lock::init_protection_request(MDL_request *mdl_request,
+                                            enum_mdl_duration duration) const
+{
+  DBUG_ENTER("Global_backup_lock::init_protection_request");
+
+  mdl_request->init(m_namespace, "", "", MDL_INTENTION_EXCLUSIVE,
+                    duration);
+
+  DBUG_VOID_RETURN;
+}
+
+/**
+   Release a protection request against a global backup lock.
+
+   @param thd  Reference to connection.
+*/
+
+void Global_backup_lock::release_protection(THD *thd)
+{
+  DBUG_ENTER("Global_backup_lock::release_protection");
+
+  DBUG_ASSERT(m_prot_lock != NULL &&
+              thd->mdl_context.is_lock_owner(m_namespace, "", "",
+                                             MDL_INTENTION_EXCLUSIVE));
+
+  thd->mdl_context.release_lock(m_prot_lock);
+
+  m_prot_lock= NULL;
+
+  DBUG_VOID_RETURN;
+}
+
 /**
   @} (end of group Locking)
 */
