@@ -1,4 +1,4 @@
-/* Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2010, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,13 +18,12 @@
 #include "sql_class.h"   // THD
 #include "sql_base.h"    // open_and_lock_tables
 #include "sql_table.h"   // write_bin_log
-#include "sql_handler.h" // mysql_ha_rm_tables
 #include "datadict.h"    // dd_recreate_table()
-#include "lock.h"        // MYSQL_OPEN_TEMPORARY_ONLY
+#include "lock.h"        // MYSQL_OPEN_* flags
 #include "sql_acl.h"     // DROP_ACL
 #include "sql_parse.h"   // check_one_table_access()
 #include "sql_truncate.h"
-#include "sql_show.h"
+#include "sql_show.h"    //append_identifier()
 #ifdef WITH_WSREP
 #include "wsrep_mysqld.h"
 #endif /* WITH_WSREP */
@@ -39,8 +38,7 @@
   @return TRUE on failure, FALSE otherwise.
 */
 
-static bool fk_info_append_fields(THD *thd, String *str,
-                                  List<LEX_STRING> *fields)
+static bool fk_info_append_fields(String *str, List<LEX_STRING> *fields)
 {
   bool res= FALSE;
   LEX_STRING *field;
@@ -48,7 +46,7 @@ static bool fk_info_append_fields(THD *thd, String *str,
 
   while ((field= it++))
   {
-    res|= append_identifier(thd, str, field->str, field->length);
+    append_identifier(NULL, str, field->str, field->length);
     res|= str->append(", ");
   }
 
@@ -80,24 +78,24 @@ static const char *fk_info_str(THD *thd, FOREIGN_KEY_INFO *fk_info)
     `db`.`tbl`, CONSTRAINT `id` FOREIGN KEY (`fk`) REFERENCES `db`.`tbl` (`fk`)
   */
 
-  res|= append_identifier(thd, &str, fk_info->foreign_db->str,
-                          fk_info->foreign_db->length);
+  append_identifier(NULL, &str, fk_info->foreign_db->str,
+                    fk_info->foreign_db->length);
   res|= str.append(".");
-  res|= append_identifier(thd, &str, fk_info->foreign_table->str,
-                          fk_info->foreign_table->length);
+  append_identifier(NULL, &str, fk_info->foreign_table->str,
+                    fk_info->foreign_table->length);
   res|= str.append(", CONSTRAINT ");
-  res|= append_identifier(thd, &str, fk_info->foreign_id->str,
-                          fk_info->foreign_id->length);
+  append_identifier(NULL, &str, fk_info->foreign_id->str,
+                    fk_info->foreign_id->length);
   res|= str.append(" FOREIGN KEY (");
-  res|= fk_info_append_fields(thd, &str, &fk_info->foreign_fields);
+  res|= fk_info_append_fields(&str, &fk_info->foreign_fields);
   res|= str.append(") REFERENCES ");
-  res|= append_identifier(thd, &str, fk_info->referenced_db->str,
-                          fk_info->referenced_db->length);
+  append_identifier(NULL, &str, fk_info->referenced_db->str,
+                    fk_info->referenced_db->length);
   res|= str.append(".");
-  res|= append_identifier(thd, &str, fk_info->referenced_table->str,
-                          fk_info->referenced_table->length);
+  append_identifier(NULL, &str, fk_info->referenced_table->str,
+                    fk_info->referenced_table->length);
   res|= str.append(" (");
-  res|= fk_info_append_fields(thd, &str, &fk_info->referenced_fields);
+  res|= fk_info_append_fields(&str, &fk_info->referenced_fields);
   res|= str.append(')');
 
   return res ? NULL : thd->strmake(str.ptr(), str.length());
@@ -185,16 +183,23 @@ fk_truncate_illegal_if_parent(THD *thd, TABLE *table)
   @param  table_ref     Table list element for the table to be truncated.
   @param  is_tmp_table  True if element refers to a temp table.
 
-  @retval  0    Success.
-  @retval  > 0  Error code.
+  @retval TRUNCATE_OK   Truncate was successful and statement can be safely
+                        binlogged.
+  @retval TRUNCATE_FAILED_BUT_BINLOG Truncate failed but still go ahead with
+                        binlogging as in case of non transactional tables
+                        partial truncation is possible.
+
+  @retval TRUNCATE_FAILED_SKIP_BINLOG Truncate was not successful hence donot
+                        binlong the statement.
 */
 
-int Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
-                                         bool is_tmp_table)
+enum Sql_cmd_truncate_table::truncate_result
+Sql_cmd_truncate_table::handler_truncate(THD *thd, TABLE_LIST *table_ref,
+                                             bool is_tmp_table)
 {
   int error= 0;
-  uint flags;
-  DBUG_ENTER("Truncate_statement::handler_truncate");
+  uint flags= 0;
+  DBUG_ENTER("Sql_cmd_truncate_table::handler_truncate");
 
   /*
     Can't recreate, the engine must mechanically delete all rows
@@ -202,9 +207,7 @@ int Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
   */
 
   /* If it is a temporary table, no need to take locks. */
-  if (is_tmp_table)
-    flags= MYSQL_OPEN_TEMPORARY_ONLY;
-  else
+  if (!is_tmp_table)
   {
     /* We don't need to load triggers. */
     DBUG_ASSERT(table_ref->trg_event_map == 0);
@@ -219,7 +222,7 @@ int Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
       the MDL lock taken above and otherwise there is no way to
       wait for FLUSH TABLES in deadlock-free fashion.
     */
-    flags= MYSQL_OPEN_IGNORE_FLUSH | MYSQL_OPEN_SKIP_TEMPORARY;
+    flags= MYSQL_OPEN_IGNORE_FLUSH;
     /*
       Even though we have an MDL lock on the table here, we don't
       pass MYSQL_OPEN_HAS_MDL_LOCK to open_and_lock_tables
@@ -232,16 +235,30 @@ int Truncate_statement::handler_truncate(THD *thd, TABLE_LIST *table_ref,
 
   /* Open the table as it will handle some required preparations. */
   if (open_and_lock_tables(thd, table_ref, FALSE, flags))
-    DBUG_RETURN(1);
+    DBUG_RETURN(TRUNCATE_FAILED_SKIP_BINLOG);
 
   /* Whether to truncate regardless of foreign keys. */
   if (! (thd->variables.option_bits & OPTION_NO_FOREIGN_KEY_CHECKS))
-    error= fk_truncate_illegal_if_parent(thd, table_ref->table);
+    if (fk_truncate_illegal_if_parent(thd, table_ref->table))
+      DBUG_RETURN(TRUNCATE_FAILED_SKIP_BINLOG);
 
-  if (!error && (error= table_ref->table->file->ha_truncate()))
+  error= table_ref->table->file->ha_truncate();
+  if (error)
+  {
     table_ref->table->file->print_error(error, MYF(0));
-
-  DBUG_RETURN(error);
+    /*
+      If truncate method is not implemented then we don't binlog the
+      statement. If truncation has failed in a transactional engine then also we
+      donot binlog the statment. Only in non transactional engine we binlog
+      inspite of errors.
+     */
+    if (error == HA_ERR_WRONG_COMMAND ||
+        table_ref->table->file->has_transactions())
+      DBUG_RETURN(TRUNCATE_FAILED_SKIP_BINLOG);
+    else
+      DBUG_RETURN(TRUNCATE_FAILED_BUT_BINLOG);
+  }
+  DBUG_RETURN(TRUNCATE_OK);
 }
 
 
@@ -269,6 +286,12 @@ static bool recreate_temporary_table(THD *thd, TABLE *table)
 
   table->file->info(HA_STATUS_AUTO | HA_STATUS_NO_LOCK);
 
+  /*
+    If LOCK TABLES list is not empty and contains this table
+    then unlock the table and remove it from this list.
+  */
+  mysql_lock_remove(thd, thd->lock, table);
+
   /* Don't free share. */
   close_temporary_table(thd, table, FALSE, FALSE);
 
@@ -278,10 +301,10 @@ static bool recreate_temporary_table(THD *thd, TABLE *table)
     on table and schema names.
   */
   ha_create_table(thd, share->normalized_path.str, share->db.str,
-                  share->table_name.str, &create_info, 1);
+                  share->table_name.str, &create_info, 0, true, true);
 
   if (open_table_uncached(thd, share->path.str, share->db.str,
-                          share->table_name.str, TRUE))
+                          share->table_name.str, true, true))
   {
     error= FALSE;
     thd->thread_specific_used= TRUE;
@@ -309,11 +332,11 @@ static bool recreate_temporary_table(THD *thd, TABLE *table)
   @retval  TRUE   Error.
 */
 
-bool Truncate_statement::lock_table(THD *thd, TABLE_LIST *table_ref,
-                                    bool *hton_can_recreate)
+bool Sql_cmd_truncate_table::lock_table(THD *thd, TABLE_LIST *table_ref,
+                                        bool *hton_can_recreate)
 {
   TABLE *table= NULL;
-  DBUG_ENTER("Truncate_statement::lock_table");
+  DBUG_ENTER("Sql_cmd_truncate_table::lock_table");
 
   /* Lock types are set in the parser. */
   DBUG_ASSERT(table_ref->lock_type == TL_WRITE);
@@ -348,8 +371,7 @@ bool Truncate_statement::lock_table(THD *thd, TABLE_LIST *table_ref,
     /* Acquire an exclusive lock. */
     DBUG_ASSERT(table_ref->next_global == NULL);
     if (lock_table_names(thd, table_ref, NULL,
-                         thd->variables.lock_wait_timeout,
-                         MYSQL_OPEN_SKIP_TEMPORARY))
+                         thd->variables.lock_wait_timeout, 0))
       DBUG_RETURN(TRUE);
 
     if (dd_check_storage_engine_flag(thd, table_ref->db, table_ref->table_name,
@@ -371,7 +393,7 @@ bool Truncate_statement::lock_table(THD *thd, TABLE_LIST *table_ref,
     m_ticket_downgrade= table->mdl_ticket;
     /* Close if table is going to be recreated. */
     if (*hton_can_recreate)
-      close_all_tables_for_name(thd, table->s, FALSE);
+      close_all_tables_for_name(thd, table->s, false, NULL);
   }
   else
   {
@@ -398,32 +420,40 @@ bool Truncate_statement::lock_table(THD *thd, TABLE_LIST *table_ref,
   @retval  TRUE   Error.
 */
 
-bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
+bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
 {
   int error;
-  TABLE *table;
   bool binlog_stmt;
-  DBUG_ENTER("Truncate_statement::truncate_table");
+  DBUG_ENTER("Sql_cmd_truncate_table::truncate_table");
+
+  DBUG_ASSERT((!table_ref->table) ||
+              (table_ref->table && table_ref->table->s));
 
   /* Initialize, or reinitialize in case of reexecution (SP). */
   m_ticket_downgrade= NULL;
 
-  /* Remove table from the HANDLER's hash. */
-  mysql_ha_rm_tables(thd, table_ref);
-
   /* If it is a temporary table, no need to take locks. */
-  if ((table= find_temporary_table(thd, table_ref)))
+  if (is_temporary_table(table_ref))
   {
+    TABLE *tmp_table= table_ref->table;
+    /*    
+      bug 1313901 : THD::decide_logging_format has not yet been called and
+                    may not be called at all dependig on the engine, so call it
+                    here.
+    */
+    if (thd->decide_logging_format(table_ref) != 0)
+      DBUG_RETURN(TRUE);
+
     /* In RBR, the statement is not binlogged if the table is temporary. */
     binlog_stmt= !thd->is_current_stmt_binlog_format_row();
 
     /* Note that a temporary table cannot be partitioned. */
-    if (ha_check_storage_engine_flag(table->s->db_type(), HTON_CAN_RECREATE))
+    if (ha_check_storage_engine_flag(tmp_table->s->db_type(), HTON_CAN_RECREATE))
     {
-      if ((error= recreate_temporary_table(thd, table)))
+      if ((error= recreate_temporary_table(thd, tmp_table)))
         binlog_stmt= FALSE; /* No need to binlog failed truncate-by-recreate. */
 
-      DBUG_ASSERT(! thd->transaction.stmt.modified_non_trans_table);
+      DBUG_ASSERT(! thd->transaction.stmt.cannot_safely_rollback());
     }
     else
     {
@@ -480,10 +510,14 @@ bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
 
       /*
         All effects of a TRUNCATE TABLE operation are committed even if
-        truncation fails. Thus, the query must be written to the binary
-        log. The only exception is a unimplemented truncate method.
+        truncation fails in the case of non transactional tables. Thus, the
+        query must be written to the binary log. The only exception is a
+        unimplemented truncate method.
       */
-      binlog_stmt= !error || error != HA_ERR_WRONG_COMMAND;
+      if (error == TRUNCATE_OK || error == TRUNCATE_FAILED_BUT_BINLOG)
+        binlog_stmt= true;
+      else
+        binlog_stmt= false;
     }
 
     /*
@@ -506,7 +540,7 @@ bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
     to a shared one.
   */
   if (m_ticket_downgrade)
-    m_ticket_downgrade->downgrade_exclusive_lock(MDL_SHARED_NO_READ_WRITE);
+    m_ticket_downgrade->downgrade_lock(MDL_SHARED_NO_READ_WRITE);
 
   DBUG_RETURN(error);
 }
@@ -519,12 +553,11 @@ bool Truncate_statement::truncate_table(THD *thd, TABLE_LIST *table_ref)
 
   @return FALSE on success.
 */
-
-bool Truncate_statement::execute(THD *thd)
+bool Sql_cmd_truncate_table::execute(THD *thd)
 {
   bool res= TRUE;
   TABLE_LIST *first_table= thd->lex->select_lex.table_list.first;
-  DBUG_ENTER("Truncate_statement::execute");
+  DBUG_ENTER("Sql_cmd_truncate_table::execute");
 
   if (check_one_table_access(thd, DROP_ACL, first_table))
     DBUG_RETURN(res);
