@@ -11,7 +11,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA */
 
 #include <my_global.h>
 #include <violite.h>
@@ -25,6 +25,7 @@
 #include <time.h>
 #include <sql_plist.h>
 #include <threadpool.h>
+#include <global_threads.h>
 #include <mysql/thread_pool_priv.h>             // thd_is_transaction_active()
 #include <time.h>
 #ifdef __linux__
@@ -250,7 +251,7 @@ static int io_poll_create()
 }
 
 
-int io_poll_associate_fd(int pollfd, int fd, void *data)
+static int io_poll_associate_fd(int pollfd, int fd, void *data)
 {
   struct epoll_event ev;
   ev.data.u64= 0; /* Keep valgrind happy */
@@ -261,7 +262,7 @@ int io_poll_associate_fd(int pollfd, int fd, void *data)
 
 
 
-int io_poll_start_read(int pollfd, int fd, void *data)
+static int io_poll_start_read(int pollfd, int fd, void *data)
 {
   struct epoll_event ev;
   ev.data.u64= 0; /* Keep valgrind happy */
@@ -270,7 +271,7 @@ int io_poll_start_read(int pollfd, int fd, void *data)
   return epoll_ctl(pollfd, EPOLL_CTL_MOD,  fd, &ev); 
 }
 
-int io_poll_disassociate_fd(int pollfd, int fd)
+static int io_poll_disassociate_fd(int pollfd, int fd)
 {
   struct epoll_event ev;
   return epoll_ctl(pollfd, EPOLL_CTL_DEL,  fd, &ev);
@@ -282,8 +283,8 @@ int io_poll_disassociate_fd(int pollfd, int fd)
  NOTE - in case of EINTR, it restarts with original timeout. Since we use
  either infinite or 0 timeouts, this is not critical
 */
-int io_poll_wait(int pollfd, native_event *native_events, int maxevents, 
-              int timeout_ms)
+static int io_poll_wait(int pollfd, native_event *native_events, int maxevents,
+                        int timeout_ms)
 {
   int ret;
   do 
@@ -301,12 +302,12 @@ static void *native_event_get_userdata(native_event *event)
 }
 
 #elif defined (__FreeBSD__) || defined (__APPLE__)
-int io_poll_create()
+static int io_poll_create()
 {
   return kqueue();
 }
 
-int io_poll_start_read(int pollfd, int fd, void *data)
+static int io_poll_start_read(int pollfd, int fd, void *data)
 {
   struct kevent ke;
   EV_SET(&ke, fd, EVFILT_READ, EV_ADD|EV_ONESHOT, 
@@ -315,7 +316,7 @@ int io_poll_start_read(int pollfd, int fd, void *data)
 }
 
 
-int io_poll_associate_fd(int pollfd, int fd, void *data)
+static int io_poll_associate_fd(int pollfd, int fd, void *data)
 {
   struct kevent ke;
   EV_SET(&ke, fd, EVFILT_READ, EV_ADD|EV_ONESHOT, 
@@ -324,7 +325,7 @@ int io_poll_associate_fd(int pollfd, int fd, void *data)
 }
 
 
-int io_poll_disassociate_fd(int pollfd, int fd)
+static int io_poll_disassociate_fd(int pollfd, int fd)
 {
   struct kevent ke;
   EV_SET(&ke,fd, EVFILT_READ, EV_DELETE, 0, 0, NULL);
@@ -332,7 +333,8 @@ int io_poll_disassociate_fd(int pollfd, int fd)
 }
 
 
-int io_poll_wait(int pollfd, struct kevent *events, int maxevents, int timeout_ms)
+static int io_poll_wait(int pollfd, struct kevent *events, int maxevents,
+                        int timeout_ms)
 {
   struct timespec ts;
   int ret;
@@ -362,7 +364,7 @@ static int io_poll_create()
   return port_create();
 }
 
-int io_poll_start_read(int pollfd, int fd, void *data)
+static int io_poll_start_read(int pollfd, int fd, void *data)
 {
   return port_associate(pollfd, PORT_SOURCE_FD, fd, POLLIN, data);
 }
@@ -372,12 +374,13 @@ static int io_poll_associate_fd(int pollfd, int fd, void *data)
   return io_poll_start_read(pollfd, fd, data);
 }
 
-int io_poll_disassociate_fd(int pollfd, int fd)
+static int io_poll_disassociate_fd(int pollfd, int fd)
 {
   return port_dissociate(pollfd, PORT_SOURCE_FD, fd);
 }
 
-int io_poll_wait(int pollfd, native_event *events, int maxevents, int timeout_ms)
+static int io_poll_wait(int pollfd, native_event *events, int maxevents,
+                        int timeout_ms)
 {
   struct timespec ts;
   int ret;
@@ -434,16 +437,23 @@ inline bool too_many_busy_threads(thread_group_t *thread_group)
 /*
    Checks if a given connection is eligible to enter the high priority queue
    based on its current thread_pool_high_prio_mode value, available high
-   priority tickets and transactional state.
+   priority tickets and transactional state and whether any locks are held.
 */
 
-inline bool connection_is_high_prio(connection_t *c)
+inline bool connection_is_high_prio(const connection_t *c)
 {
   const ulong mode= c->thd->variables.threadpool_high_prio_mode;
 
   return (mode == TP_HIGH_PRIO_MODE_STATEMENTS) ||
-    (mode == TP_HIGH_PRIO_MODE_TRANSACTIONS &&
-     c->tickets > 0 && thd_is_transaction_active(c->thd));
+    (mode == TP_HIGH_PRIO_MODE_TRANSACTIONS && c->tickets > 0 &&
+     (thd_is_transaction_active(c->thd) ||
+      c->thd->variables.option_bits & OPTION_TABLE_LOCK ||
+      c->thd->locked_tables_mode != LTM_NONE ||
+      c->thd->mdl_context.has_locks() ||
+      c->thd->global_read_lock.is_acquired() ||
+      c->thd->backup_tables_lock.is_acquired() ||
+      c->thd->backup_binlog_lock.is_acquired() ||
+      c->thd->ull_hash.records > 0));
 }
 
 } // namespace
@@ -481,17 +491,22 @@ static connection_t *queue_get(thread_group_t *thread_group)
 
 static void timeout_check(pool_timer_t *timer)
 {
+  std::set<THD*> global_thread_list_copy;
   DBUG_ENTER("timeout_check");
-  
-  mysql_mutex_lock(&LOCK_thread_count);
-  I_List_iterator<THD> it(threads);
+
+  mysql_mutex_lock(&LOCK_thd_remove);
+  copy_global_thread_list(&global_thread_list_copy);
+
+  Thread_iterator it= global_thread_list_copy.begin();
+  Thread_iterator end= global_thread_list_copy.end();
 
   /* Reset next timeout check, it will be recalculated in the loop below */
   my_atomic_fas64((volatile int64*)&timer->next_timeout_check, ULONGLONG_MAX);
 
   THD *thd;
-  while ((thd=it++))
+  for ( ; it != end; ++it)
   {
+    thd= (*it);
     if (thd->net.reading_or_writing != 1)
       continue;
  
@@ -518,7 +533,7 @@ static void timeout_check(pool_timer_t *timer)
       set_next_timeout_check(connection->abs_wait_timeout);
     }
   }
-  mysql_mutex_unlock(&LOCK_thread_count);
+  mysql_mutex_unlock(&LOCK_thd_remove);
   DBUG_VOID_RETURN;
 }
 
@@ -970,7 +985,8 @@ static int wake_or_create_thread(thread_group_t *thread_group)
 
 
 
-int thread_group_init(thread_group_t *thread_group, pthread_attr_t* thread_attr)
+static int thread_group_init(thread_group_t *thread_group,
+                             pthread_attr_t* thread_attr)
 {
   DBUG_ENTER("thread_group_init");
   thread_group->pthread_attr = thread_attr;
@@ -982,7 +998,7 @@ int thread_group_init(thread_group_t *thread_group, pthread_attr_t* thread_attr)
 }
 
 
-void thread_group_destroy(thread_group_t *thread_group)
+static void thread_group_destroy(thread_group_t *thread_group)
 {
   mysql_mutex_destroy(&thread_group->mutex);
   if (thread_group->pollfd != -1)
@@ -1110,8 +1126,9 @@ static void queue_put(thread_group_t *thread_group, connection_t *connection)
   NULL is returned if timeout has expired,or on shutdown.
 */
 
-connection_t *get_event(worker_thread_t *current_thread, 
-  thread_group_t *thread_group,  struct timespec *abstime)
+static connection_t *get_event(worker_thread_t *current_thread,
+                               thread_group_t *thread_group,
+                               struct timespec *abstime)
 { 
   DBUG_ENTER("get_event");
   connection_t *connection = NULL;
@@ -1241,7 +1258,7 @@ connection_t *get_event(worker_thread_t *current_thread,
   sleep() or similar.
 */
 
-void wait_begin(thread_group_t *thread_group)
+static void wait_begin(thread_group_t *thread_group)
 {
   DBUG_ENTER("wait_begin");
   mysql_mutex_lock(&thread_group->mutex);
@@ -1271,7 +1288,7 @@ void wait_begin(thread_group_t *thread_group)
   Tells the pool has finished waiting.
 */
 
-void wait_end(thread_group_t *thread_group)
+static void wait_end(thread_group_t *thread_group)
 {
   DBUG_ENTER("wait_end");
   mysql_mutex_lock(&thread_group->mutex);
@@ -1286,7 +1303,7 @@ void wait_end(thread_group_t *thread_group)
   Allocate/initialize a new connection structure.
 */
 
-connection_t *alloc_connection(THD *thd)
+static connection_t *alloc_connection(THD *thd)
 {
   DBUG_ENTER("alloc_connection");
   
@@ -1313,7 +1330,7 @@ void tp_add_connection(THD *thd)
 {
   DBUG_ENTER("tp_add_connection");
   
-  threads.append(thd);
+  add_global_thread(thd);
   mysql_mutex_unlock(&LOCK_thread_count);
   connection_t *connection= alloc_connection(thd);
   if (connection)
@@ -1376,7 +1393,7 @@ void tp_post_kill_notification(THD *thd)
     DBUG_VOID_RETURN;
   
   if (thd->net.vio)
-    vio_shutdown(thd->net.vio, SHUT_RD);
+    vio_cancel(thd->net.vio, SHUT_RD);
   DBUG_VOID_RETURN;
 }
 
@@ -1468,7 +1485,7 @@ static int change_group(connection_t *c,
  thread_group_t *new_group)
 { 
   int ret= 0;
-  int fd = c->thd->net.vio->sd;
+  int fd = mysql_socket_getfd(c->thd->net.vio->mysql_socket);
 
   DBUG_ASSERT(c->thread_group == old_group);
 
@@ -1496,7 +1513,7 @@ static int change_group(connection_t *c,
 
 static int start_io(connection_t *connection)
 { 
-  int fd = connection->thd->net.vio->sd;
+  int fd = mysql_socket_getfd(connection->thd->net.vio->mysql_socket);
 
   /*
     Usually, connection will stay in the same group for the entire
@@ -1582,6 +1599,11 @@ static void *worker_main(void *param)
   this_thread.thread_group= thread_group;
   this_thread.event_count=0;
 
+#ifdef HAVE_PSI_THREAD_INTERFACE
+    PSI_THREAD_CALL(set_thread_user_host)
+      (NULL, 0, NULL, 0);
+#endif
+
   /* Run event loop */
   for(;;)
   {
@@ -1751,26 +1773,29 @@ A likely cause of pool blocks are clients that lock resources for long time. \
 
 static void print_pool_blocked_message(bool max_threads_reached)
 {
-  ulonglong now;
+  ulonglong now= my_microsecond_getsystime();
   static bool msg_written;
-  
-  now= my_microsecond_getsystime();
+
   if (pool_block_start == 0)
   {
     pool_block_start= now;
-    msg_written = false;
-    return;
+    msg_written= false;
   }
-  
-  if (now > pool_block_start + BLOCK_MSG_DELAY && !msg_written)
+
+  if (!msg_written
+      && ((now > pool_block_start + BLOCK_MSG_DELAY)
+          || (now == pool_block_start)))
   {
     if (max_threads_reached)
       sql_print_error(MAX_THREADS_REACHED_MSG);
     else
       sql_print_error(CREATE_THREAD_ERROR_MSG, my_errno);
-    
-    sql_print_information("Threadpool has been blocked for %u seconds\n",
-      (uint)((now- pool_block_start)/1000000));
+
+    if (now > pool_block_start)
+    {
+      sql_print_information("Threadpool has been blocked for %u seconds\n",
+                            (uint)((now - pool_block_start)/1000000));
+    }
     /* avoid reperated messages for the same blocking situation */
     msg_written= true;
   }
