@@ -1,4 +1,4 @@
-/* Copyright 2008 Codership Oy <http://www.codership.com>
+/* Copyright 2008-2015 Codership Oy <http://www.codership.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,19 +11,20 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
 
 #include <mysqld.h>
 #include "sql_base.h"
+#include "binlog.h"
 #include "rpl_filter.h"
 #include <sql_class.h>
 #include "wsrep_mysqld.h"
 #include "wsrep_binlog.h"
+#include "wsrep_xid.h"
 #include <cstdio>
 #include <cstdlib>
+#include "debug_sync.h"
 
-extern handlerton *binlog_hton;
-extern int binlog_close_connection(handlerton *hton, THD *thd);
 extern ulonglong thd_to_trx_id(THD *thd);
 
 extern "C" int thd_binlog_format(const MYSQL_THD thd);
@@ -37,11 +38,15 @@ enum wsrep_trx_status wsrep_run_wsrep_commit(THD *thd, handlerton *hton,
 */
 void wsrep_cleanup_transaction(THD *thd)
 {
+  if (!WSREP(thd)) return;
+
   if (wsrep_emulate_bin_log) thd_binlog_trx_reset(thd);
   thd->wsrep_ws_handle.trx_id= WSREP_UNDEFINED_TRX_ID;
   thd->wsrep_trx_meta.gtid= WSREP_GTID_UNDEFINED;
   thd->wsrep_trx_meta.depends_on= WSREP_SEQNO_UNDEFINED;
   thd->wsrep_exec_mode= LOCAL_STATE;
+  thd->wsrep_certify_empty_trx= false;
+  thd->wsrep_affected_rows= 0;
   return;
 }
 
@@ -68,20 +73,31 @@ handlerton *wsrep_hton;
 */
 void wsrep_register_hton(THD* thd, bool all)
 {
-  if (thd->wsrep_exec_mode != TOTAL_ORDER && !thd->wsrep_apply_toi)
+  if (WSREP(thd) && thd->wsrep_exec_mode != TOTAL_ORDER && !thd->wsrep_apply_toi)
   {
+    if (thd->wsrep_exec_mode == LOCAL_STATE      &&
+        (thd_sql_command(thd) == SQLCOM_OPTIMIZE ||
+        thd_sql_command(thd) == SQLCOM_ANALYZE   ||
+        thd_sql_command(thd) == SQLCOM_REPAIR)   &&
+            thd->lex->no_write_to_binlog == 1)
+    {
+      WSREP_DEBUG("Skipping wsrep_register_hton for LOCAL admin command: %s",
+                  WSREP_QUERY(thd));
+        return;
+    }
+
     THD_TRANS *trans=all ? &thd->transaction.all : &thd->transaction.stmt;
-    for (Ha_trx_info *i= trans->ha_list; WSREP(thd) && i; i = i->next())
+    for (Ha_trx_info *i= trans->ha_list; i; i = i->next())
     {
       if (i->ht()->db_type == DB_TYPE_INNODB)
       {
         trans_register_ha(thd, all, wsrep_hton);
 
         /* follow innodb read/write settting
-         * but, as an exception: CTAS with empty result set will not be 
+         * but, as an exception: CTAS with empty result set will not be
          * replicated unless we declare wsrep hton as read/write here
 	 */
-        if (i->is_trx_read_write() || 
+        if (i->is_trx_read_write() ||
             (thd->lex->sql_command == SQLCOM_CREATE_TABLE &&
              thd->wsrep_exec_mode == LOCAL_STATE))
         {
@@ -99,17 +115,35 @@ void wsrep_register_hton(THD* thd, bool all)
  */
 void wsrep_post_commit(THD* thd, bool all)
 {
-  if (thd->wsrep_exec_mode == LOCAL_COMMIT)
+  if (!WSREP(thd)) return;
+
+  switch (thd->wsrep_exec_mode)
   {
-    DBUG_ASSERT(thd->wsrep_trx_meta.gtid.seqno != WSREP_SEQNO_UNDEFINED);
-    if (wsrep->post_commit(wsrep, &thd->wsrep_ws_handle))
+  case LOCAL_COMMIT:
     {
+      DBUG_ASSERT(thd->wsrep_trx_meta.gtid.seqno != WSREP_SEQNO_UNDEFINED);
+      if (wsrep->post_commit(wsrep, &thd->wsrep_ws_handle))
+      {
         DBUG_PRINT("wsrep", ("set committed fail"));
-        WSREP_WARN("set committed fail: %llu %d", 
-                   (long long)thd->real_id, thd->stmt_da->status());
+        WSREP_WARN("set committed fail: %llu %d",
+                   (long long)thd->real_id, thd->get_stmt_da()->status());
+      }
+      wsrep_cleanup_transaction(thd);
+      break;
     }
-    wsrep_cleanup_transaction(thd);
+ case LOCAL_STATE:
+   {
+     /* non-InnoDB statements may have populated events in stmt cache 
+	=> cleanup 
+     */
+     WSREP_DEBUG("cleanup transaction for LOCAL_STATE: %s",
+                 WSREP_QUERY(thd));
+     wsrep_cleanup_transaction(thd);
+     break;
+   }
+  default: break;
   }
+
 }
 
 /*
@@ -128,10 +162,7 @@ wsrep_close_connection(handlerton*  hton, THD* thd)
   {
     DBUG_RETURN(0);
   }
-
-  if (wsrep_emulate_bin_log && thd_get_ha_data(thd, binlog_hton) != NULL)
-    binlog_hton->close_connection (binlog_hton, thd);
-  DBUG_RETURN(0);
+  DBUG_RETURN(wsrep_binlog_close_connection (thd));
 }
 
 /*
@@ -174,10 +205,11 @@ static int wsrep_savepoint_set(handlerton *hton, THD *thd,  void *sv)
     DBUG_RETURN(0);
   }
 
-  if (!wsrep_emulate_bin_log) return 0;
-  int rcode = binlog_hton->savepoint_set(binlog_hton, thd, sv);
-  return rcode;
+  if (!wsrep_emulate_bin_log) DBUG_RETURN(0);
+  int rcode = wsrep_binlog_savepoint_set(thd, sv);
+  DBUG_RETURN(rcode);
 }
+
 static int wsrep_savepoint_rollback(handlerton *hton, THD *thd, void *sv)
 {
   DBUG_ENTER("wsrep_savepoint_rollback");
@@ -187,9 +219,9 @@ static int wsrep_savepoint_rollback(handlerton *hton, THD *thd, void *sv)
     DBUG_RETURN(0);
   }
 
-  if (!wsrep_emulate_bin_log) return 0;
-  int rcode = binlog_hton->savepoint_rollback(binlog_hton, thd, sv);
-  return rcode;
+  if (!wsrep_emulate_bin_log) DBUG_RETURN(0);
+  int rcode = wsrep_binlog_savepoint_rollback(thd, sv);
+  DBUG_RETURN(rcode);
 }
 
 static int wsrep_rollback(handlerton *hton, THD *thd, bool all)
@@ -202,14 +234,26 @@ static int wsrep_rollback(handlerton *hton, THD *thd, bool all)
   }
 
   mysql_mutex_lock(&thd->LOCK_wsrep_thd);
+  switch (thd->wsrep_exec_mode)
+  {
+  case TOTAL_ORDER:
+  case REPL_RECV:
+      mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
+      WSREP_DEBUG("Avoiding wsrep rollback for failed DDL: %s", 
+                  WSREP_QUERY(thd));
+      DBUG_RETURN(0);
+  default: break;
+  }
+
   if ((all || !thd_test_options(thd, OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN)) &&
       (thd->variables.wsrep_on && thd->wsrep_conflict_state != MUST_REPLAY))
   {
     if (wsrep->post_rollback(wsrep, &thd->wsrep_ws_handle))
     {
       DBUG_PRINT("wsrep", ("setting rollback fail"));
-      WSREP_ERROR("settting rollback fail: thd: %llu SQL: %s",
-		  (long long)thd->real_id, thd->query());
+      WSREP_ERROR("settting rollback fail: thd: %llu, schema: %s, SQL: %s",
+                  (long long)thd->real_id, (thd->db ? thd->db : "(null)"),
+                  WSREP_QUERY(thd));
     }
     wsrep_cleanup_transaction(thd);
   }
@@ -248,8 +292,9 @@ int wsrep_commit(handlerton *hton, THD *thd, bool all)
       if (wsrep->post_rollback(wsrep, &thd->wsrep_ws_handle))
       {
         DBUG_PRINT("wsrep", ("setting rollback fail"));
-        WSREP_ERROR("settting rollback fail: thd: %llu SQL: %s", 
-                    (long long)thd->real_id, thd->query());
+        WSREP_ERROR("settting rollback fail: thd: %llu, schema: %s, SQL: %s",
+                    (long long)thd->real_id, (thd->db ? thd->db : "(null)"),
+                    WSREP_QUERY(thd));
       }
       wsrep_cleanup_transaction(thd);
     }
@@ -270,12 +315,15 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   IO_CACHE *cache;
   int replay_round= 0;
 
-  if (thd->stmt_da->is_error()) {
+  if (thd->get_stmt_da()->is_error()) {
     WSREP_ERROR("commit issue, error: %d %s",
-                thd->stmt_da->sql_errno(), thd->stmt_da->message());
+                thd->get_stmt_da()->sql_errno(), thd->get_stmt_da()->message());
   }
 
   DBUG_ENTER("wsrep_run_wsrep_commit");
+#ifdef WITH_WSREP
+  DEBUG_SYNC(thd, "wsrep_before_replication");
+#endif /* WITH_WSREP */
 
   if (thd->slave_thread && !opt_log_slave_updates) DBUG_RETURN(WSREP_TRX_OK);
 
@@ -300,7 +348,7 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   if (thd->wsrep_exec_mode != LOCAL_STATE) DBUG_RETURN(WSREP_TRX_OK);
 
   if (thd->wsrep_consistency_check == CONSISTENCY_CHECK_RUNNING) {
-    WSREP_DEBUG("commit for consistency check: %s", thd->query());
+    WSREP_DEBUG("commit for consistency check: %s", WSREP_QUERY(thd));
     DBUG_RETURN(WSREP_TRX_OK);
   }
 
@@ -311,10 +359,7 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
     DBUG_PRINT("wsrep", ("replicate commit fail"));
     thd->wsrep_conflict_state = ABORTED;
     mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
-    if (wsrep_debug) {
-      WSREP_INFO("innobase_commit, abort %s",
-                 (thd->query()) ? thd->query() : "void");
-    }
+    WSREP_DEBUG("innobase_commit, abort %s", WSREP_QUERY(thd));
     DBUG_RETURN(WSREP_TRX_CERT_FAIL);
   }
 
@@ -338,7 +383,12 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
     mysql_mutex_lock(&LOCK_wsrep_replaying);
     // Using timedwait is a hack to avoid deadlock in case if BF victim
     // misses the signal.
-    struct timespec wtime = {0, 1000000};
+    struct timespec wtime;
+    clock_gettime(CLOCK_REALTIME, &wtime);
+    long prev_nsec = wtime.tv_nsec;
+    wtime.tv_nsec = (wtime.tv_nsec + 1000000) % 1000000000;
+    // If nsecs rolled over, increment seconds.
+    wtime.tv_sec += (wtime.tv_nsec < prev_nsec ? 1 : 0);
     mysql_cond_timedwait(&COND_wsrep_replaying, &LOCK_wsrep_replaying,
 			 &wtime);
 
@@ -365,7 +415,7 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
     thd->wsrep_conflict_state = ABORTED;
     mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
     WSREP_DEBUG("innobase_commit abort after replaying wait %s",
-                (thd->query()) ? thd->query() : "void");
+                WSREP_QUERY(thd));
     DBUG_RETURN(WSREP_TRX_CERT_FAIL);
   }
 
@@ -385,23 +435,25 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
 
   if (data_len == 0)
   {
-    if (thd->stmt_da->is_ok()              &&
-        thd->stmt_da->affected_rows() > 0  &&
+    if (thd->get_stmt_da()->is_ok()              &&
+        thd->get_stmt_da()->affected_rows() > 0  &&
         !binlog_filter->is_on())
     {
       WSREP_DEBUG("empty rbr buffer, query: %s, "
-                 "affected rows: %llu, "
-                 "changed tables: %d, "
-                 "sql_log_bin: %d, "
-                 "wsrep status (%d %d %d)",
-                 thd->query(), thd->stmt_da->affected_rows(),
-                 stmt_has_updated_trans_table(thd), thd->variables.sql_log_bin,
-                 thd->wsrep_exec_mode, thd->wsrep_query_state,
-                 thd->wsrep_conflict_state);
+                  "affected rows: %llu, "
+                  "changed tables: %d, "
+                  "sql_log_bin: %d, "
+                  "wsrep status (%d %d %d)",
+                  WSREP_QUERY(thd),
+                  thd->get_stmt_da()->affected_rows(),
+                  stmt_has_updated_trans_table(thd->transaction.stmt.ha_list),
+                  thd->variables.sql_log_bin,
+                  thd->wsrep_exec_mode, thd->wsrep_query_state,
+                  thd->wsrep_conflict_state);
     }
     else
     {
-      WSREP_DEBUG("empty rbr buffer, query: %s", thd->query());
+      WSREP_DEBUG("empty rbr buffer, query: %s", WSREP_QUERY(thd));
     }
     thd->wsrep_query_state= QUERY_EXEC;
     DBUG_RETURN(WSREP_TRX_OK);
@@ -410,34 +462,39 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   if (WSREP_UNDEFINED_TRX_ID == thd->wsrep_ws_handle.trx_id)
   {
     WSREP_WARN("SQL statement was ineffective, THD: %lu, buf: %zu\n"
-               "QUERY: %s\n"
-               " => Skipping replication", 
-               thd->thread_id, data_len, thd->query());
+               "schema: %s \n"
+	       "QUERY: %s\n"
+	       " => Skipping replication",
+	       thd->thread_id, data_len,
+               (thd->db ? thd->db : "(null)"),
+               WSREP_QUERY(thd));
     rcode = WSREP_TRX_FAIL;
   }
   else if (!rcode)
   {
     if (WSREP_OK == rcode) {
-        if (WSREP(thd))
-            thd_proc_info(thd, "wsrep in pre-commit stage");
-        rcode = wsrep->pre_commit(wsrep,
-                                    (wsrep_conn_id_t)thd->thread_id,
-                                    &thd->wsrep_ws_handle,
-                                    WSREP_FLAG_COMMIT |
-                                    ((thd->wsrep_PA_safe) ?
-                                    0ULL : WSREP_FLAG_PA_UNSAFE),
-                                    &thd->wsrep_trx_meta);
+      if (WSREP(thd))
+        thd_proc_info(thd, "wsrep in pre-commit stage");
+      rcode = wsrep->pre_commit(wsrep,
+                                (wsrep_conn_id_t)thd->thread_id,
+                                &thd->wsrep_ws_handle,
+                                WSREP_FLAG_COMMIT |
+                                ((thd->wsrep_PA_safe) ?
+                                 0ULL : WSREP_FLAG_PA_UNSAFE),
+                                &thd->wsrep_trx_meta);
     }
 
     if (rcode == WSREP_TRX_MISSING) {
-      WSREP_WARN("Transaction missing in provider, thd: %ld, SQL: %s",
-                 thd->thread_id, thd->query());
+      WSREP_WARN("Transaction missing in provider thd: %ld schema: %s SQL: %s",
+                 thd->thread_id, (thd->db ? thd->db : "(null)"),
+                 WSREP_QUERY(thd));
       rcode = WSREP_TRX_FAIL;
     } else if (rcode == WSREP_BF_ABORT) {
       WSREP_DEBUG("thd %lu seqno %lld BF aborted by provider, will replay",
                   thd->thread_id, (long long)thd->wsrep_trx_meta.gtid.seqno);
       mysql_mutex_lock(&thd->LOCK_wsrep_thd);
       thd->wsrep_conflict_state = MUST_REPLAY;
+      DBUG_ASSERT(wsrep_thd_trx_seqno(thd) > 0);
       mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
       mysql_mutex_lock(&LOCK_wsrep_replaying);
       wsrep_replaying++;
@@ -477,7 +534,7 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
     if (thd->transaction.xid_state.xid.get_my_xid())
     {
       wsrep_xid_init(&thd->transaction.xid_state.xid,
-                     &thd->wsrep_trx_meta.gtid.uuid,
+                     thd->wsrep_trx_meta.gtid.uuid,
                      thd->wsrep_trx_meta.gtid.seqno);
     }
     DBUG_PRINT("wsrep", ("replicating commit success"));
@@ -485,7 +542,8 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   case WSREP_BF_ABORT:
     DBUG_ASSERT(thd->wsrep_trx_meta.gtid.seqno != WSREP_SEQNO_UNDEFINED);
   case WSREP_TRX_FAIL:
-    WSREP_DEBUG("commit failed for reason: %d", rcode);
+    WSREP_DEBUG("commit failed for reason: %d %lu %s",
+                rcode, thd->thread_id, WSREP_QUERY(thd));
     DBUG_PRINT("wsrep", ("replicating commit fail"));
 
     thd->wsrep_query_state= QUERY_EXEC;
