@@ -1,4 +1,4 @@
-/* Copyright 2008-2012 Codership Oy <http://www.codership.com>
+/* Copyright 2008-2015 Codership Oy <http://www.codership.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
 
 #include "wsrep_sst.h"
 
@@ -23,17 +23,21 @@
 #include <sql_parse.h>
 #include "wsrep_priv.h"
 #include "wsrep_utils.h"
+#include "wsrep_xid.h"
 #include <cstdio>
 #include <cstdlib>
 
 extern const char wsrep_defaults_file[];
+extern const char wsrep_defaults_group_suffix[];
 
 #define WSREP_SST_OPT_ROLE     "--role"
 #define WSREP_SST_OPT_ADDR     "--address"
 #define WSREP_SST_OPT_AUTH     "--auth"
 #define WSREP_SST_OPT_DATA     "--datadir"
 #define WSREP_SST_OPT_CONF     "--defaults-file"
+#define WSREP_SST_OPT_CONF_SUFFIX "--defaults-group-suffix"
 #define WSREP_SST_OPT_PARENT   "--parent"
+#define WSREP_SST_OPT_BINLOG   "--binlog"
 
 // mysqldump-specific options
 #define WSREP_SST_OPT_USER     "--user"
@@ -47,9 +51,9 @@ extern const char wsrep_defaults_file[];
 #define WSREP_SST_OPT_GTID     "--gtid"
 #define WSREP_SST_OPT_BYPASS   "--bypass"
 
-#define WSREP_SST_MYSQLDUMP    "mysqldump"
-#define WSREP_SST_RSYNC        "rsync"
-#define WSREP_SST_SKIP         "skip"
+#define WSREP_SST_MYSQLDUMP       "mysqldump"
+#define WSREP_SST_RSYNC           "rsync"
+#define WSREP_SST_SKIP            "skip"
 #define WSREP_SST_XTRABACKUP      "xtrabackup"
 #define WSREP_SST_XTRABACKUP_V2   "xtrabackup-v2"
 #define WSREP_SST_DEFAULT      WSREP_SST_XTRABACKUP_V2
@@ -65,19 +69,20 @@ const char* wsrep_sst_donor           = "";
 static const char* sst_auth_real      = NULL;
 my_bool wsrep_sst_donor_rejects_queries = FALSE;
 
+/* Function checks if the new value for sst_method is valid.
+@return false if no error encountered with check else return true. */
 bool wsrep_sst_method_check (sys_var *self, THD* thd, set_var* var)
 {
-    char   buff[FN_REFLEN];
-    String str(buff, sizeof(buff), system_charset_info), *res;
-    const char* c_str = NULL;
+  if ((! var->save_result.string_value.str) ||
+      (var->save_result.string_value.length == 0 ))
+  {
+    my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+             var->save_result.string_value.str ?
+             var->save_result.string_value.str : "NULL");
+    return true;
+  }
 
-    if ((res   = var->value->val_str(&str)) &&
-        (c_str = res->c_ptr()) &&
-        strlen(c_str) > 0)
-        return 0;
-
-    my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), "wsrep_sst_method", c_str ? c_str : "NULL");
-    return 1;
+  return false;
 }
 
 bool wsrep_sst_method_update (sys_var *self, THD* thd, enum_var_type type)
@@ -96,20 +101,37 @@ static bool sst_receive_address_check (const char* str)
     return 0;
 }
 
-bool  wsrep_sst_receive_address_check (sys_var *self, THD* thd, set_var* var)
+/* Function checks if the new value for sst_recieve_address is valid.
+@return false if no error encountered with check else return true. */
+bool wsrep_sst_receive_address_check (sys_var *self, THD* thd, set_var* var)
 {
-    const char* c_str = var->value->str_value.c_ptr();
+  char addr_buf[FN_REFLEN];
 
-    if (sst_receive_address_check (c_str))
-    {
-        my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), "wsrep_sst_receive_address", c_str ? c_str : "NULL");
-        return 1;
-    }
+  if ((! var->save_result.string_value.str) ||
+      (var->save_result.string_value.length > (FN_REFLEN - 1))) // safety
+  {
+    goto err;
+  }
 
-    return 0;
+  memcpy(addr_buf, var->save_result.string_value.str,
+         var->save_result.string_value.length);
+  addr_buf[var->save_result.string_value.length]= 0;
+
+  if (sst_receive_address_check(addr_buf))
+  {
+    goto err;
+  }
+
+  return false;
+
+err:
+  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+           var->save_result.string_value.str ?
+           var->save_result.string_value.str : "NULL");
+  return true;
 }
 
-bool wsrep_sst_receive_address_update (sys_var *self, THD* thd, 
+bool wsrep_sst_receive_address_update (sys_var *self, THD* thd,
                                        enum_var_type type)
 {
     return 0;
@@ -119,31 +141,44 @@ bool wsrep_sst_auth_check (sys_var *self, THD* thd, set_var* var)
 {
     return 0;
 }
+
+void wsrep_sst_auth_free()
+{
+  if (wsrep_sst_auth) { my_free((void *) wsrep_sst_auth); }
+  if (sst_auth_real) { my_free((void *) sst_auth_real); }
+  wsrep_sst_auth= NULL;
+  sst_auth_real= NULL;
+}
+
 static bool sst_auth_real_set (const char* value)
 {
-    const char* v = strdup (value);
+  const char* v= NULL;
 
-    if (v)
+  if (value)
+  {
+    v= my_strdup(value, MYF(0));
+  }
+  else                                          // its NULL
+  {
+    wsrep_sst_auth_free();
+    return 0;
+  }
+
+  if (v)
+  {
+    // set sst_auth_real
+    if (sst_auth_real) { my_free((void *) sst_auth_real); }
+    sst_auth_real = v;
+
+    // mask wsrep_sst_auth
+    if (strlen(sst_auth_real))
     {
-        if (sst_auth_real) free (const_cast<char*>(sst_auth_real));
-        sst_auth_real = v;
-
-        if (strlen(sst_auth_real))
-        {
-          if (wsrep_sst_auth)
-          {
-            my_free ((void*)wsrep_sst_auth);
-            wsrep_sst_auth = my_strdup(WSREP_SST_AUTH_MASK, MYF(0));
-            //strncpy (wsrep_sst_auth, WSREP_SST_AUTH_MASK,
-            //     sizeof(wsrep_sst_auth) - 1);
-          }
-          else
-            wsrep_sst_auth = my_strdup (WSREP_SST_AUTH_MASK, MYF(0));
-        }
-        return 0;
+      if (wsrep_sst_auth) { my_free((void*) wsrep_sst_auth); }
+      wsrep_sst_auth= my_strdup(WSREP_SST_AUTH_MASK, MYF(0));
     }
-
-    return 1;
+    return 0;
+  }
+  return 1;
 }
 
 bool wsrep_sst_auth_update (sys_var *self, THD* thd, enum_var_type type)
@@ -191,6 +226,8 @@ void wsrep_sst_grab ()
 // Wait for end of SST
 bool wsrep_sst_wait ()
 {
+  wsrep_seqno_t seqno;
+
   if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
   while (!sst_complete)
   {
@@ -198,19 +235,26 @@ bool wsrep_sst_wait ()
     mysql_cond_wait (&COND_wsrep_sst, &LOCK_wsrep_sst);
   }
 
-  if (local_seqno >= 0)
+  // Save a local copy of a global variable prior to release
+  // of the mutex - to ensure that the return code from this
+  // function and a diagnostic message in the log will always
+  // be consistent with each other:
+
+  seqno = local_seqno;
+
+  mysql_mutex_unlock (&LOCK_wsrep_sst);
+
+  if (seqno >= 0)
   {
-    WSREP_INFO("SST complete, seqno: %lld", (long long) local_seqno);
+    WSREP_INFO("SST complete, seqno: %lld", (long long) seqno);
   }
   else
   {
     WSREP_ERROR("SST failed: %d (%s)",
-                int(-local_seqno), strerror(-local_seqno));
+                int(-seqno), strerror(-seqno));
   }
 
-  mysql_mutex_unlock (&LOCK_wsrep_sst);
-
-  return (local_seqno >= 0);
+  return (seqno >= 0);
 }
 
 // Signal end of SST
@@ -240,17 +284,46 @@ void wsrep_sst_complete (const wsrep_uuid_t* sst_uuid,
   mysql_mutex_unlock (&LOCK_wsrep_sst);
 }
 
+// True if wsrep awaiting sst_received callback:
+
+static bool sst_awaiting_callback= false;
+
 void wsrep_sst_received (wsrep_t*            const wsrep,
-                         const wsrep_uuid_t* const uuid,
+                         const wsrep_uuid_t& uuid,
                          wsrep_seqno_t       const seqno,
                          const void*         const state,
                          size_t              const state_len)
 {
-    int const rcode(seqno < 0 ? seqno : 0);
-    wsrep_gtid_t const state_id = {
-        *uuid, (rcode ? WSREP_SEQNO_UNDEFINED : seqno)
-    };
-    wsrep->sst_received(wsrep, &state_id, state, state_len, rcode);
+    wsrep_get_SE_checkpoint(local_uuid, local_seqno);
+
+    if (memcmp(&local_uuid, &uuid, sizeof(wsrep_uuid_t)) ||
+        local_seqno < seqno || seqno < 0)
+    {
+        wsrep_set_SE_checkpoint(uuid, seqno);
+        local_uuid = uuid;
+        local_seqno = seqno;
+    }
+    else if (local_seqno > seqno)
+    {
+        WSREP_WARN("SST postion is in the past: %lld, current: %lld. "
+                   "Can't continue.",
+                   (long long)seqno, (long long)local_seqno);
+        unireg_abort(1);
+    }
+
+    wsrep_init_sidno(uuid);
+
+    if (wsrep)
+    {
+        int const rcode(seqno < 0 ? seqno : 0);
+        wsrep_gtid_t const state_id = {
+            uuid, (rcode ? WSREP_SEQNO_UNDEFINED : seqno)
+        };
+        if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+        sst_awaiting_callback = false;
+        wsrep->sst_received(wsrep, &state_id, state, state_len, rcode);
+        mysql_mutex_unlock (&LOCK_wsrep_sst);
+    }
 }
 
 // Let applier threads to continue
@@ -259,19 +332,21 @@ void wsrep_sst_continue ()
   if (sst_needed)
   {
     WSREP_INFO("Signalling provider to continue.");
-    wsrep_sst_received (wsrep, &local_uuid, local_seqno, NULL, 0);
+    wsrep_sst_received (wsrep, local_uuid, local_seqno, NULL, 0);
   }
 }
 
 struct sst_thread_arg
 {
   const char*     cmd;
-  int             err;
+  char**          env;
   char*           ret_str;
+  int             err;
   mysql_mutex_t   lock;
   mysql_cond_t    cond;
 
-  sst_thread_arg (const char* c) : cmd(c), err(-1), ret_str(0)
+  sst_thread_arg (const char* c, char** e)
+    : cmd(c), env(e), ret_str(0), err(-1)
   {
     mysql_mutex_init(key_LOCK_wsrep_sst_thread, &lock, MY_MUTEX_INIT_FAST);
     mysql_cond_init(key_COND_wsrep_sst_thread, &cond, NULL);
@@ -316,6 +391,94 @@ static char* my_fgets (char* buf, size_t buf_len, FILE* stream)
    return ret;
 }
 
+/*
+  Generate opt_binlog_opt_val for sst_donate_other(), sst_prepare_other().
+
+  Returns zero on success, negative error code otherwise.
+
+  String containing binlog name is stored in param ret if binlog is enabled
+  and GTID mode is on, otherwise empty string. Returned string should be
+  freed with my_free().
+ */
+static int generate_binlog_opt_val(char** ret)
+{
+  DBUG_ASSERT(ret);
+  *ret= NULL;
+  if (opt_bin_log && gtid_mode > 0)
+  {
+    assert(opt_bin_logname);
+    *ret= strcmp(opt_bin_logname, "0") ?
+        my_strdup(opt_bin_logname, MYF(0)) : my_strdup("", MYF(0));
+  }
+  else
+  {
+    *ret= my_strdup("", MYF(0));
+  }
+  if (!*ret) return -ENOMEM;
+  return 0;
+}
+
+// Pointer to SST process object:
+
+static wsp::process * sst_process= static_cast<wsp::process*>(NULL);
+
+// SST cancellation flag:
+
+static bool sst_cancelled= false;
+
+void wsrep_sst_cancel (bool call_wsrep_cb)
+{
+  if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+  if (!sst_cancelled)
+  {
+    sst_cancelled=true;
+    /*
+      When we launched the SST process, then we need
+      to terminate it before exit from the parent (server)
+      process:
+    */
+    if (sst_process)
+    {
+      sst_process->terminate();
+      sst_process = NULL;
+    }
+    /*
+      If this is a normal shutdown, then we need to notify
+      the wsrep provider about completion of the SST, to
+      prevent infinite waitng in the wsrep provider after
+      the SST process was canceled:
+    */
+    if (call_wsrep_cb && sst_awaiting_callback)
+    {
+      WSREP_INFO("Signalling cancellation of the SST request.");
+      wsrep_gtid_t const state_id = {
+          WSREP_UUID_UNDEFINED,
+          WSREP_SEQNO_UNDEFINED
+      };
+      sst_awaiting_callback = false;
+      wsrep->sst_received(wsrep, &state_id, NULL, 0, -ECANCELED);
+    }
+  }
+  mysql_mutex_unlock (&LOCK_wsrep_sst);
+}
+
+/*
+  Handling of fatal signals: SIGABRT, SIGTERM, etc.
+*/
+extern "C" void wsrep_handle_fatal_signal (int sig)
+{
+  wsrep_sst_cancel(false);
+}
+
+/*
+  This callback is invoked in the case of abnormal
+  termination of the wsrep provider:
+*/
+void wsrep_abort_cb (void)
+{
+  wsrep_sst_cancel(false);
+}
+
 static void* sst_joiner_thread (void* a)
 {
   sst_thread_arg* arg= (sst_thread_arg*) a;
@@ -329,7 +492,12 @@ static void* sst_joiner_thread (void* a)
 
     WSREP_INFO("Running: '%s'", arg->cmd);
 
-    wsp::process proc (arg->cmd, "r");
+    // Launch the SST script and save pointer to its process:
+
+    if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+    wsp::process proc (arg->cmd, "r", arg->env);
+    sst_process = &proc;
+    mysql_mutex_unlock (&LOCK_wsrep_sst);
 
     if (proc.pipe() && !proc.error())
     {
@@ -338,18 +506,36 @@ static void* sst_joiner_thread (void* a)
       if (!tmp || strlen(tmp) < (magic_len + 2) ||
           strncasecmp (tmp, magic, magic_len))
       {
-        WSREP_ERROR("Failed to read '%s <addr>' from: %s\n\tRead: '%s'",
-                    magic, arg->cmd, tmp);
+        if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+        // Print error message if SST is not cancelled:
+        if (!sst_cancelled)
+        {
+          // Null-pointer is not valid argument for %s formatting (even
+          // though it is supported by many compilers):
+           WSREP_ERROR("Failed to read '%s <addr>' from: %s\n\tRead: '%s'",
+                       magic, arg->cmd, tmp ? tmp : "(null)");
+        }
+        // Clear the pointer to SST process:
+        sst_process = NULL;
+        mysql_mutex_unlock (&LOCK_wsrep_sst);
         proc.wait();
         if (proc.error()) err = proc.error();
       }
       else
       {
+        // Clear the pointer to SST process:
+        if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+        sst_process = NULL;
+        mysql_mutex_unlock (&LOCK_wsrep_sst);
         err = 0;
       }
     }
     else
     {
+      // Clear the pointer to SST process:
+      if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+      sst_process = NULL;
+      mysql_mutex_unlock (&LOCK_wsrep_sst);
       err = proc.error();
       WSREP_ERROR("Failed to execute: %s : %d (%s)",
                   arg->cmd, err, strerror(err));
@@ -380,10 +566,15 @@ static void* sst_joiner_thread (void* a)
     proc.wait();
     err= EINVAL;
 
-    if (!tmp)
+    if (!tmp || proc.error())
     {
       WSREP_ERROR("Failed to read uuid:seqno from joiner script.");
-      if (proc.error()) err = proc.error();
+      if (proc.error())
+      {
+        err= proc.error();
+        char errbuf[MYSYS_STRERROR_SIZE];
+        WSREP_ERROR("SST script aborted with error %d (%s)", err, my_strerror(errbuf, sizeof(errbuf), err));
+      }
     }
     else
     {
@@ -403,24 +594,69 @@ static void* sst_joiner_thread (void* a)
   return NULL;
 }
 
+#define WSREP_SST_AUTH_ENV "WSREP_SST_OPT_AUTH"
+
+static int sst_append_auth_env(wsp::env& env, const char* sst_auth)
+{
+  int const sst_auth_size= strlen(WSREP_SST_AUTH_ENV) + 1 /* = */
+    + (sst_auth ? strlen(sst_auth) : 0) + 1 /* \0 */;
+
+  wsp::string sst_auth_str(sst_auth_size); // for automatic cleanup on return
+  if (!sst_auth_str()) return -ENOMEM;
+
+  int ret= snprintf(sst_auth_str(), sst_auth_size, "%s=%s",
+                    WSREP_SST_AUTH_ENV, sst_auth ? sst_auth : "");
+
+  if (ret < 0 || ret >= sst_auth_size)
+  {
+    WSREP_ERROR("sst_append_auth_env(): snprintf() failed: %d", ret);
+    return (ret < 0 ? ret : -EMSGSIZE);
+  }
+
+  env.append(sst_auth_str());
+  return -env.error();
+}
+
 static ssize_t sst_prepare_other (const char*  method,
+                                  const char*  sst_auth,
                                   const char*  addr_in,
                                   const char** addr_out)
 {
-  ssize_t cmd_len= 1024;
-  char    cmd_str[cmd_len];
-  const char* sst_dir= mysql_real_data_home;
+  int const cmd_len= 4096;
+  wsp::string cmd_str(cmd_len);
 
-  int ret= snprintf (cmd_str, cmd_len,
-                     "wsrep_sst_%s "
-                     WSREP_SST_OPT_ROLE" 'joiner' "
-                     WSREP_SST_OPT_ADDR" '%s' "
-                     WSREP_SST_OPT_AUTH" '%s' "
-                     WSREP_SST_OPT_DATA" '%s' "
-                     WSREP_SST_OPT_CONF" '%s' "
-                     WSREP_SST_OPT_PARENT" '%d'",
-                     method, addr_in, (sst_auth_real) ? sst_auth_real : "",
-                     sst_dir, wsrep_defaults_file, (int)getpid());
+  if (!cmd_str())
+  {
+    WSREP_ERROR("sst_prepare_other(): could not allocate cmd buffer of %d bytes",
+                cmd_len);
+    return -ENOMEM;
+  }
+
+  const char* binlog_opt= "";
+  char* binlog_opt_val= NULL;
+
+  int ret;
+  if ((ret= generate_binlog_opt_val(&binlog_opt_val)))
+  {
+    WSREP_ERROR("sst_prepare_other(): generate_binlog_opt_val() failed: %d",
+                ret);
+    return ret;
+  }
+  if (strlen(binlog_opt_val)) binlog_opt= WSREP_SST_OPT_BINLOG;
+
+  ret= snprintf (cmd_str(), cmd_len,
+                 "wsrep_sst_%s "
+                 WSREP_SST_OPT_ROLE" 'joiner' "
+                 WSREP_SST_OPT_ADDR" '%s' "
+                 WSREP_SST_OPT_DATA" '%s' "
+                 WSREP_SST_OPT_CONF" '%s' "
+                 WSREP_SST_OPT_CONF_SUFFIX" '%s' "
+                 WSREP_SST_OPT_PARENT" '%d'"
+                 " %s '%s' ",
+                 method, addr_in, mysql_real_data_home,
+                 wsrep_defaults_file, wsrep_defaults_group_suffix,
+                 (int)getpid(), binlog_opt, binlog_opt_val);
+  my_free(binlog_opt_val);
 
   if (ret < 0 || ret >= cmd_len)
   {
@@ -428,15 +664,28 @@ static ssize_t sst_prepare_other (const char*  method,
     return (ret < 0 ? ret : -EMSGSIZE);
   }
 
+  wsp::env env(NULL);
+  if (env.error())
+  {
+    WSREP_ERROR("sst_prepare_other(): env. var ctor failed: %d", -env.error());
+    return -env.error();
+  }
+
+  if ((ret= sst_append_auth_env(env, sst_auth)))
+  {
+    WSREP_ERROR("sst_prepare_other(): appending auth failed: %d", ret);
+    return ret;
+  }
+
   pthread_t tmp;
-  sst_thread_arg arg(cmd_str);
+  sst_thread_arg arg(cmd_str(), env());
   mysql_mutex_lock (&arg.lock);
   ret = pthread_create (&tmp, NULL, sst_joiner_thread, &arg);
   if (ret)
   {
     WSREP_ERROR("sst_prepare_other(): pthread_create() failed: %d (%s)",
                 ret, strerror(ret));
-    return ret;
+    return -ret;
   }
   mysql_cond_wait (&arg.cond, &arg.lock);
 
@@ -455,7 +704,6 @@ static ssize_t sst_prepare_other (const char*  method,
   return ret;
 }
 
-//extern ulong my_bind_addr;
 extern uint  mysqld_port;
 
 /*! Just tells donor where to send mysqldump */
@@ -579,7 +827,8 @@ ssize_t wsrep_sst_prepare (void** msg)
       return 0;
     }
 
-    addr_len = sst_prepare_other (wsrep_sst_method, addr_in, &addr_out);
+    addr_len = sst_prepare_other (wsrep_sst_method, sst_auth_real,
+                                  addr_in, &addr_out);
     if (addr_len < 0)
     {
       WSREP_ERROR("Failed to prepare for '%s' SST. Unrecoverable.",
@@ -599,6 +848,10 @@ ssize_t wsrep_sst_prepare (void** msg)
     strcpy (addr_ptr, addr_out);
 
     WSREP_INFO ("Prepared SST request: %s|%s", method_ptr, addr_ptr);
+
+    if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+    sst_awaiting_callback = true;
+    mysql_mutex_unlock (&LOCK_wsrep_sst);
   }
   else {
     WSREP_ERROR("Failed to allocate SST request of size %zu. Can't continue.",
@@ -612,24 +865,45 @@ ssize_t wsrep_sst_prepare (void** msg)
 }
 
 // helper method for donors
-static int sst_run_shell (const char* cmd_str, int max_tries)
+static int sst_run_shell (const char* cmd_str, char** env, int max_tries)
 {
   int ret = 0;
 
   for (int tries=1; tries <= max_tries; tries++)
   {
-    wsp::process proc (cmd_str, "r");
+    // Launch the SST script and save pointer to its process:
 
-    if (NULL != proc.pipe())
+    if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+    wsp::process proc (cmd_str, "r", env);
+    sst_process = &proc;
+    mysql_mutex_unlock (&LOCK_wsrep_sst);
+
+    if (proc.pipe() && !proc.error())
     {
       proc.wait();
     }
 
+    // Clear the pointer to SST process:
+
+    if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+    sst_process = NULL;
+    mysql_mutex_unlock (&LOCK_wsrep_sst);
+
     if ((ret = proc.error()))
     {
-      WSREP_ERROR("Try %d/%d: '%s' failed: %d (%s)",
-                  tries, max_tries, proc.cmd(), ret, strerror(ret));
-      sleep (1);
+      // Try again if SST is not cancelled:
+      if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+      if (!sst_cancelled)
+      {
+         mysql_mutex_unlock (&LOCK_wsrep_sst);
+         WSREP_ERROR("Try %d/%d: '%s' failed: %d (%s)",
+                     tries, max_tries, proc.cmd(), ret, strerror(ret));
+         sleep (1);
+      }
+      else
+      {
+         mysql_mutex_unlock (&LOCK_wsrep_sst);
+      }
     }
     else
     {
@@ -648,17 +922,12 @@ static void sst_reject_queries(my_bool close_conn)
     if (TRUE == close_conn) wsrep_close_client_connections(FALSE);
 }
 
-static int sst_mysqldump_check_addr (const char* user, const char* pswd,
-                                     const char* host, const char* port)
-{
-  return 0;
-}
-
 static int sst_donate_mysqldump (const char*         addr,
                                  const wsrep_uuid_t* uuid,
                                  const char*         uuid_str,
                                  wsrep_seqno_t       seqno,
-                                 bool                bypass)
+                                 bool                bypass,
+                                 char**              env) // carries auth info
 {
   size_t host_len;
   const char* port = strchr (addr, ':');
@@ -674,64 +943,51 @@ static int sst_donate_mysqldump (const char*         addr,
     host_len = strlen (addr) + 1;
   }
 
-  char host[host_len];
+  char *host= static_cast<char *>(alloca(host_len));
 
   strncpy (host, addr, host_len - 1);
   host[host_len - 1] = '\0';
 
-  const char* auth = sst_auth_real;
-  const char* pswd = (auth) ? strchr (auth, ':') : NULL;
-  size_t user_len;
+  int const cmd_len= 4096;
+  wsp::string  cmd_str(cmd_len);
 
-  if (pswd)
+  if (!cmd_str())
   {
-    pswd += 1;
-    user_len = pswd - auth;
-  }
-  else
-  {
-    pswd = "";
-    user_len = (auth) ? strlen (auth) + 1 : 1;
+    WSREP_ERROR("sst_donate_mysqldump(): "
+                "could not allocate cmd buffer of %d bytes", cmd_len);
+    return -ENOMEM;
   }
 
-  char user[user_len];
+  if (!bypass && wsrep_sst_donor_rejects_queries) sst_reject_queries(TRUE);
 
-  strncpy (user, (auth) ? auth : "", user_len - 1);
-  user[user_len - 1] = '\0';
+  int ret= snprintf (cmd_str(), cmd_len,
+                     "wsrep_sst_mysqldump "
+                     WSREP_SST_OPT_HOST" '%s' "
+                     WSREP_SST_OPT_PORT" '%s' "
+                     WSREP_SST_OPT_LPORT" '%u' "
+                     WSREP_SST_OPT_SOCKET" '%s' "
+                     WSREP_SST_OPT_CONF" '%s' "
+                     WSREP_SST_OPT_GTID" '%s:%lld'"
+                     "%s",
+                     host, port, mysqld_port, mysqld_unix_port,
+                     wsrep_defaults_file, uuid_str,
+                     (long long)seqno, bypass ? " " WSREP_SST_OPT_BYPASS : "");
 
-  int ret = sst_mysqldump_check_addr (user, pswd, host, port);
-  if (!ret)
+  if (ret < 0 || ret >= cmd_len)
   {
-    size_t cmd_len= 1024;
-    char   cmd_str[cmd_len];
-
-    if (!bypass && wsrep_sst_donor_rejects_queries) sst_reject_queries(TRUE);
-
-    snprintf (cmd_str, cmd_len,
-              "wsrep_sst_mysqldump "
-              WSREP_SST_OPT_USER" '%s' "
-              WSREP_SST_OPT_PSWD" '%s' "
-              WSREP_SST_OPT_HOST" '%s' "
-              WSREP_SST_OPT_PORT" '%s' "
-              WSREP_SST_OPT_LPORT" '%u' "
-              WSREP_SST_OPT_SOCKET" '%s' "
-              WSREP_SST_OPT_DATA" '%s' "
-              WSREP_SST_OPT_CONF" '%s' "
-              WSREP_SST_OPT_GTID" '%s:%lld'"
-              "%s",
-              user, pswd, host, port, mysqld_port, mysqld_unix_port,
-              mysql_real_data_home, wsrep_defaults_file, uuid_str,
-              (long long)seqno, bypass ? " "WSREP_SST_OPT_BYPASS : "");
-
-    WSREP_DEBUG("Running: '%s'", cmd_str);
-
-    ret= sst_run_shell (cmd_str, 3);
+    WSREP_ERROR("sst_donate_mysqldump(): snprintf() failed: %d", ret);
+    return (ret < 0 ? ret : -EMSGSIZE);
   }
 
+  WSREP_DEBUG("Running: '%s'", cmd_str());
+
+  ret= sst_run_shell (cmd_str(), env, 3);
+
   wsrep_gtid_t const state_id = { *uuid, (ret ? WSREP_SEQNO_UNDEFINED : seqno)};
 
   wsrep->sst_sent (wsrep, &state_id, ret);
 
+
   return ret;
 }
 
@@ -751,10 +1007,10 @@ static int run_sql_command(THD *thd, const char *query)
   mysql_parse(thd, thd->query(), thd->query_length(), &ps);
   if (thd->is_error())
   {
-    int const err= thd->stmt_da->sql_errno();
+    int const err= thd->get_stmt_da()->sql_errno();
     WSREP_WARN ("error executing '%s': %d (%s)%s",
-                query, err, thd->stmt_da->message(),
-                err == ER_UNKNOWN_SYSTEM_VARIABLE ? 
+                query, err, thd->get_stmt_da()->message(),
+                err == ER_UNKNOWN_SYSTEM_VARIABLE ?
                 ". Was mysqld built with --with-innodb-disallow-writes ?" : "");
     thd->clear_error();
     return -1;
@@ -771,13 +1027,13 @@ static int sst_flush_tables(THD* thd)
   if (run_sql_command(thd, "FLUSH TABLES WITH READ LOCK"))
   {
     WSREP_ERROR("Failed to flush and lock tables");
-    err = -1;
+    err = ECANCELED;
   }
   else
   {
     /* make sure logs are flushed after global read lock acquired */
-    err= reload_acl_and_cache(thd, REFRESH_ENGINE_LOG,
-                              (TABLE_LIST*) 0, &not_used);
+    err= reload_acl_and_cache(thd, REFRESH_ENGINE_LOG | REFRESH_BINARY_LOG,
+			      (TABLE_LIST*) 0, &not_used);
   }
 
   if (err)
@@ -789,9 +1045,9 @@ static int sst_flush_tables(THD* thd)
     WSREP_INFO("Tables flushed.");
     const char base_name[]= "tables_flushed";
     ssize_t const full_len= strlen(mysql_real_data_home) + strlen(base_name)+2;
-    char real_name[full_len];
+    char *real_name= static_cast<char *>(alloca(full_len));
     sprintf(real_name, "%s/%s", mysql_real_data_home, base_name);
-    char tmp_name[full_len + 4];
+    char *tmp_name= static_cast<char *>(alloca(full_len + 4));
     sprintf(tmp_name, "%s.tmp", real_name);
 
     FILE* file= fopen(tmp_name, "w+");
@@ -849,7 +1105,14 @@ static void* sst_donor_thread (void* a)
 
   wsp::thd thd(FALSE); // we turn off wsrep_on for this THD so that it can
                        // operate with wsrep_ready == OFF
-  wsp::process proc(arg->cmd, "r");
+  thd.ptr->wsrep_sst_donor= true;
+
+  // Launch the SST script and save pointer to its process:
+
+  if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+  wsp::process proc(arg->cmd, "r", arg->env);
+  sst_process = &proc;
+  mysql_mutex_unlock (&LOCK_wsrep_sst);
 
   err= proc.error();
 
@@ -903,13 +1166,33 @@ wait_signal:
     }
     else
     {
-      WSREP_ERROR("Failed to read from: %s", proc.cmd());
+      if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+      // Print error message if SST is not cancelled:
+      if (!sst_cancelled)
+      {
+         WSREP_ERROR("Failed to read from: %s", proc.cmd());
+      }
+      // Clear the pointer to SST process:
+      sst_process = NULL;
+      mysql_mutex_unlock (&LOCK_wsrep_sst);
       proc.wait();
+      goto skip_clear_pointer;
     }
+    
+    // Clear the pointer to SST process:
+    if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+    sst_process = NULL;
+    mysql_mutex_unlock (&LOCK_wsrep_sst);
+
+skip_clear_pointer:
     if (!err && proc.error()) err= proc.error();
   }
   else
   {
+    // Clear the pointer to SST process:
+    if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+    sst_process = NULL;
+    mysql_mutex_unlock (&LOCK_wsrep_sst);
     WSREP_ERROR("Failed to execute: %s : %d (%s)",
                 proc.cmd(), err, strerror(err));
   }
@@ -930,29 +1213,53 @@ wait_signal:
   return NULL;
 }
 
+
+
 static int sst_donate_other (const char*   method,
                              const char*   addr,
                              const char*   uuid,
                              wsrep_seqno_t seqno,
-                             bool          bypass)
+                             bool          bypass,
+                             char**        env) // carries auth info
 {
-  ssize_t cmd_len = 4096;
-  char    cmd_str[cmd_len];
+  int const cmd_len= 4096;
+  wsp::string  cmd_str(cmd_len);
 
-  int ret= snprintf (cmd_str, cmd_len,
-                     "wsrep_sst_%s "
-                     WSREP_SST_OPT_ROLE" 'donor' "
-                     WSREP_SST_OPT_ADDR" '%s' "
-                     WSREP_SST_OPT_AUTH" '%s' "
-                     WSREP_SST_OPT_SOCKET" '%s' "
-                     WSREP_SST_OPT_DATA" '%s' "
-                     WSREP_SST_OPT_CONF" '%s' "
-                     WSREP_SST_OPT_GTID" '%s:%lld'"
-                     "%s",
-                     method, addr, sst_auth_real, mysqld_unix_port,
-                     mysql_real_data_home, wsrep_defaults_file,
-                     uuid, (long long) seqno,
-                     bypass ? " "WSREP_SST_OPT_BYPASS : "");
+  if (!cmd_str())
+  {
+    WSREP_ERROR("sst_donate_other(): "
+                "could not allocate cmd buffer of %d bytes", cmd_len);
+    return -ENOMEM;
+  }
+
+  const char* binlog_opt= "";
+  char* binlog_opt_val= NULL;
+
+  int ret;
+  if ((ret= generate_binlog_opt_val(&binlog_opt_val)))
+  {
+    WSREP_ERROR("sst_donate_other(): generate_binlog_opt_val() failed: %d",ret);
+    return ret;
+  }
+  if (strlen(binlog_opt_val)) binlog_opt= WSREP_SST_OPT_BINLOG;
+
+  ret= snprintf (cmd_str(), cmd_len,
+                 "wsrep_sst_%s "
+                 WSREP_SST_OPT_ROLE" 'donor' "
+                 WSREP_SST_OPT_ADDR" '%s' "
+                 WSREP_SST_OPT_SOCKET" '%s' "
+                 WSREP_SST_OPT_DATA" '%s' "
+                 WSREP_SST_OPT_CONF" '%s' "
+                 WSREP_SST_OPT_CONF_SUFFIX" '%s' "
+                 " %s '%s' "
+                 WSREP_SST_OPT_GTID" '%s:%lld'"
+                 "%s",
+                 method, addr, mysqld_unix_port, mysql_real_data_home,
+                 wsrep_defaults_file, wsrep_defaults_group_suffix,
+                 binlog_opt, binlog_opt_val,
+                 uuid, (long long) seqno,
+                 bypass ? " " WSREP_SST_OPT_BYPASS : "");
+  my_free(binlog_opt_val);
 
   if (ret < 0 || ret >= cmd_len)
   {
@@ -963,14 +1270,14 @@ static int sst_donate_other (const char*   method,
   if (!bypass && wsrep_sst_donor_rejects_queries) sst_reject_queries(FALSE);
 
   pthread_t tmp;
-  sst_thread_arg arg(cmd_str);
+  sst_thread_arg arg(cmd_str(), env);
   mysql_mutex_lock (&arg.lock);
   ret = pthread_create (&tmp, NULL, sst_donor_thread, &arg);
   if (ret)
   {
     WSREP_ERROR("sst_donate_other(): pthread_create() failed: %d (%s)",
                 ret, strerror(ret));
-    return -ret;
+    return ret;
   }
   mysql_cond_wait (&arg.cond, &arg.lock);
 
@@ -996,18 +1303,32 @@ wsrep_cb_status_t wsrep_sst_donate_cb (void* app_ctx, void* recv_ctx,
   char uuid_str[37];
   wsrep_uuid_print (&current_gtid->uuid, uuid_str, sizeof(uuid_str));
 
+  wsp::env env(NULL);
+  if (env.error())
+  {
+    WSREP_ERROR("wsrep_sst_donate_cb(): env var ctor failed: %d", -env.error());
+    return WSREP_CB_FAILURE;
+  }
+
   int ret;
+  if ((ret= sst_append_auth_env(env, sst_auth_real)))
+  {
+    WSREP_ERROR("wsrep_sst_donate_cb(): appending auth env failed: %d", ret);
+    return WSREP_CB_FAILURE;
+  }
+
   if (!strcmp (WSREP_SST_MYSQLDUMP, method))
   {
     ret = sst_donate_mysqldump(data, &current_gtid->uuid, uuid_str,
-                               current_gtid->seqno, bypass);
+                               current_gtid->seqno, bypass, env());
   }
   else
   {
-    ret = sst_donate_other(method, data, uuid_str, current_gtid->seqno,bypass);
+    ret = sst_donate_other(method, data, uuid_str,
+                           current_gtid->seqno, bypass, env());
   }
 
-  return (ret > 0 ? WSREP_CB_SUCCESS : WSREP_CB_FAILURE);
+  return (ret >= 0 ? WSREP_CB_SUCCESS : WSREP_CB_FAILURE);
 }
 
 void wsrep_SE_init_grab()
