@@ -19,7 +19,7 @@
 #include "rpl_rli.h"
 #include "log_event.h"
 #include "sql_parse.h"
-#include "slave.h"    // opt_log_slave_updates
+#include "global_threads.h" // LOCK_thread_count, etc.
 #include "sql_base.h" // close_thread_tables()
 #include "mysqld.h"   // start_wsrep_THD();
 
@@ -37,12 +37,16 @@ int wsrep_show_bf_aborts (THD *thd, SHOW_VAR *var, char *buff)
 void wsrep_client_rollback(THD *thd)
 {
   WSREP_DEBUG("client rollback due to BF abort for (%ld), query: %s",
-              thd->thread_id, thd->query());
+              thd->thread_id, WSREP_QUERY(thd));
 
   my_atomic_add64(&wsrep_bf_aborts_counter, 1);
 
   thd->wsrep_conflict_state= ABORTING;
   mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
+
+  /* Check for comments in Relay_log_info::cleanup_context */
+  trans_rollback_stmt(thd);
+
   trans_rollback(thd);
 
   if (thd->locked_tables_mode && thd->lock)
@@ -61,6 +65,9 @@ void wsrep_client_rollback(THD *thd)
   /* Release transactional metadata locks. */
   thd->mdl_context.release_transactional_locks();
 
+  /* Release all user-locks. */
+  mysql_ull_cleanup(thd);
+
   /* release explicit MDL locks */
   thd->mdl_context.release_explicit_locks();
 
@@ -73,19 +80,74 @@ void wsrep_client_rollback(THD *thd)
   thd->wsrep_conflict_state= ABORTED;
 }
 
+#define NUMBER_OF_FIELDS_TO_IDENTIFY_COORDINATOR 1
+#define NUMBER_OF_FIELDS_TO_IDENTIFY_WORKER 2
+#include "rpl_info_factory.h"
+
 static Relay_log_info* wsrep_relay_log_init(const char* log_fname)
 {
-  Relay_log_info* rli= new Relay_log_info(false);
+  uint rli_option = INFO_REPOSITORY_DUMMY;
+  Relay_log_info *rli= NULL;
+  rli = Rpl_info_factory::create_rli(rli_option, false);
+  rli->set_rli_description_event(
+      new Format_description_log_event(BINLOG_VERSION));
 
-  rli->no_storage= true;
-  if (!rli->relay_log.description_event_for_exec)
+  return (rli);
+
+#ifdef REMOVED
+  Rpl_info_handler* handler_src= NULL;
+  Rpl_info_handler* handler_dest= NULL;
+  ulong *key_info_idx= NULL;
+  const char *msg= "Failed to allocate memory for the relay log info "
+                   "structure";
+
+  DBUG_ENTER("Rpl_info_factory::create_rli");
+
+  if (!(rli= new Relay_log_info(false
+#ifdef HAVE_PSI_INTERFACE
+                                ,&key_relay_log_info_run_lock,
+                                &key_relay_log_info_data_lock,
+                                &key_relay_log_info_sleep_lock,
+                                &key_relay_log_info_data_cond,
+                                &key_relay_log_info_start_cond,
+                                &key_relay_log_info_stop_cond,
+                                &key_relay_log_info_sleep_cond
+#endif /* HAVE_PSI_INTERFACE */
+                               )))
+    goto err;
+
+  if (!(key_info_idx= new ulong[NUMBER_OF_FIELDS_TO_IDENTIFY_COORDINATOR]))
+     goto err;
+  key_info_idx[0]= server_id;
+  rli->set_idx_info(key_info_idx, NUMBER_OF_FIELDS_TO_IDENTIFY_COORDINATOR);
+
+  if(Rpl_info_factory::init_rli_repositories(rli, rli_option, &handler_src,
+                                             &handler_dest, &msg))
+    goto err;
+
+  if (Rpl_info_factory::decide_repository(rli, rli_option, &handler_src,
+                                          &handler_dest, &msg))
+    goto err;
+
+  DBUG_RETURN(rli);
+err:
+  delete handler_src;
+  delete handler_dest;
+  delete []key_info_idx;
+  if (rli)
   {
-    rli->relay_log.description_event_for_exec=
-      new Format_description_log_event(4);
+    /*
+      The handler was previously deleted so we need to remove
+      any reference to it.
+    */
+    rli->set_idx_info(NULL, 0);
+    rli->set_rpl_info_handler(NULL);
+    rli->set_rpl_info_type(INVALID_INFO_REPOSITORY);
+    delete rli;
   }
-
-  rli->sql_thd= current_thd;
-  return rli;
+  WSREP_ERROR("Error creating relay log info: %s.", msg);
+  DBUG_RETURN(NULL);
+#endif /* REMOVED */
 }
 
 static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
@@ -95,12 +157,16 @@ static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
   shadow->wsrep_exec_mode = thd->wsrep_exec_mode;
   shadow->vio           = thd->net.vio;
 
+  // Disable general logging on applier threads
+  thd->variables.option_bits |= OPTION_LOG_OFF;
+  // Enable binlogging if opt_log_slave_updates is set
   if (opt_log_slave_updates)
     thd->variables.option_bits|= OPTION_BIN_LOG;
   else
     thd->variables.option_bits&= ~(OPTION_BIN_LOG);
 
   if (!thd->wsrep_rli) thd->wsrep_rli= wsrep_relay_log_init("wsrep_relay");
+  thd->wsrep_rli->info_thd = thd;
 
   thd->wsrep_exec_mode= REPL_RECV;
   thd->net.vio= 0;
@@ -113,6 +179,8 @@ static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
   shadow->db            = thd->db;
   shadow->db_length     = thd->db_length;
   thd->reset_db(NULL, 0);
+
+  shadow->user_time = thd->user_time;
 }
 
 static void wsrep_return_from_bf_mode(THD *thd, struct wsrep_thd_shadow* shadow)
@@ -123,18 +191,30 @@ static void wsrep_return_from_bf_mode(THD *thd, struct wsrep_thd_shadow* shadow)
   thd->net.vio                = shadow->vio;
   thd->variables.tx_isolation = shadow->tx_isolation;
   thd->reset_db(shadow->db, shadow->db_length);
+  thd->user_time              = shadow->user_time;
 }
 
 void wsrep_replay_transaction(THD *thd)
 {
+  DBUG_ENTER("wsrep_replay_transaction");
   /* checking if BF trx must be replayed */
   if (thd->wsrep_conflict_state== MUST_REPLAY) {
+    DBUG_ASSERT(wsrep_thd_trx_seqno(thd));
     if (thd->wsrep_exec_mode!= REPL_RECV) {
-      if (thd->stmt_da->is_sent)
+      if (thd->get_stmt_da()->is_sent())
       {
         WSREP_ERROR("replay issue, thd has reported status already");
       }
-      thd->stmt_da->reset_diagnostics_area();
+
+      /* PS reprepare observer should have been removed already
+         open_table() will fail if we have dangling observer here
+       */
+      if (thd->get_reprepare_observer() && wsrep_log_conflicts)
+      {
+        WSREP_WARN("dangling observer in replay transaction: (thr %lu %lld %s)",
+                   thd->thread_id, thd->query_id, thd->query());
+      }
+      thd->get_stmt_da()->reset_diagnostics_area();
 
       thd->wsrep_conflict_state= REPLAYING;
       mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
@@ -150,10 +230,17 @@ void wsrep_replay_transaction(THD *thd)
         thd->variables.option_bits&= ~(OPTION_TABLE_LOCK);
       }
       thd->mdl_context.release_transactional_locks();
-
+      /*
+        Replaying will call MYSQL_START_STATEMENT when handling
+        BEGIN Query_log_event so end statement must be called before
+        replaying.
+      */
+      MYSQL_END_STATEMENT(thd->m_statement_psi, thd->get_stmt_da());
+      thd->m_statement_psi= NULL;
+      thd->m_digest= NULL;
       thd_proc_info(thd, "wsrep replaying trx");
       WSREP_DEBUG("replay trx: %s %lld",
-                  thd->query() ? thd->query() : "void",
+                  WSREP_QUERY(thd),
                   (long long)wsrep_thd_trx_seqno(thd));
       struct wsrep_thd_shadow shadow;
       wsrep_prepare_bf_thd(thd, &shadow);
@@ -179,16 +266,16 @@ void wsrep_replay_transaction(THD *thd)
         wsrep->post_commit(wsrep, &thd->wsrep_ws_handle);
         WSREP_DEBUG("trx_replay successful for: %ld %llu",
                     thd->thread_id, (long long)thd->real_id);
-        if (thd->stmt_da->is_sent)
+        if (thd->get_stmt_da()->is_sent())
         {
           WSREP_WARN("replay ok, thd has reported status");
         }
-        else if (thd->stmt_da->is_set())
+        else if (thd->get_stmt_da()->is_set())
         {
-          if (thd->stmt_da->status() != Diagnostics_area::DA_OK)
+          if (thd->get_stmt_da()->status() != Diagnostics_area::DA_OK)
           {
             WSREP_WARN("replay ok, thd has error status %d",
-                       thd->stmt_da->status());
+                       thd->get_stmt_da()->status());
           }
         }
         else
@@ -197,22 +284,23 @@ void wsrep_replay_transaction(THD *thd)
         }
         break;
       case WSREP_TRX_FAIL:
-        if (thd->stmt_da->is_sent)
+        if (thd->get_stmt_da()->is_sent())
         {
           WSREP_ERROR("replay failed, thd has reported status");
         }
         else
         {
           WSREP_DEBUG("replay failed, rolling back");
-          my_error(ER_LOCK_DEADLOCK, MYF(0), "wsrep aborted transaction");
+          //my_error(ER_LOCK_DEADLOCK, MYF(0), "wsrep aborted transaction");
         }
         thd->wsrep_conflict_state= ABORTED;
         wsrep->post_rollback(wsrep, &thd->wsrep_ws_handle);
         break;
       default:
-        WSREP_ERROR("trx_replay failed for: %d, query: %s",
-                    rcode, thd->query() ? thd->query() : "void");
+        WSREP_ERROR("trx_replay failed for: %d, schema: %s, query: %s",
+                    rcode, (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
         /* we're now in inconsistent state, must abort */
+	mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
         unireg_abort(1);
         break;
       }
@@ -227,6 +315,7 @@ void wsrep_replay_transaction(THD *thd)
       mysql_mutex_unlock(&LOCK_wsrep_replaying);
     }
   }
+  DBUG_VOID_RETURN;
 }
 
 static void wsrep_replication_process(THD *thd)
@@ -235,6 +324,7 @@ static void wsrep_replication_process(THD *thd)
   DBUG_ENTER("wsrep_replication_process");
 
   struct wsrep_thd_shadow shadow;
+
   wsrep_prepare_bf_thd(thd, &shadow);
 
   /* From trans_begin() */
@@ -286,9 +376,13 @@ static void wsrep_replication_process(THD *thd)
   mysql_cond_broadcast(&COND_thread_count);
   mysql_mutex_unlock(&LOCK_thread_count);
 
-  if (thd->temporary_tables)
+  TABLE *tmp;
+  while ((tmp = thd->temporary_tables))
   {
-    WSREP_DEBUG("Applier %lu, has temporary tables at exit", thd->thread_id);
+    WSREP_WARN("Applier %lu, has temporary tables at exit: %s.%s",
+                  thd->thread_id, 
+                  (tmp->s) ? tmp->s->db.str : "void",
+                  (tmp->s) ? tmp->s->table_name.str : "void");
   }
   wsrep_return_from_bf_mode(thd, &shadow);
   DBUG_VOID_RETURN;
@@ -302,9 +396,8 @@ void wsrep_create_appliers(long threads)
     if (wsrep_cluster_address && strlen(wsrep_cluster_address) &&
         wsrep_provider && strcasecmp(wsrep_provider, "none"))
     {
-      WSREP_ERROR("Trying to launch slave threads before creating "
+      WSREP_WARN("Trying to launch slave threads before creating "
                   "connection at '%s'", wsrep_cluster_address);
-      assert(0);
     }
     return;
   }
@@ -405,7 +498,29 @@ void wsrep_create_rollbacker()
   }
 }
 
-extern "C"
+void wsrep_thd_set_PA_safe(void *thd_ptr, my_bool safe)
+{ 
+  if (thd_ptr) 
+  {
+    THD* thd = (THD*)thd_ptr;
+    thd->wsrep_PA_safe = safe;
+  }
+}
+
+int wsrep_thd_conflict_state(void *thd_ptr, my_bool sync)
+{ 
+  int state = -1;
+  if (thd_ptr) 
+  {
+    THD* thd = (THD*)thd_ptr;
+    if (sync) mysql_mutex_lock(&thd->LOCK_wsrep_thd);
+    
+    state = thd->wsrep_conflict_state;
+    if (sync) mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
+  }
+  return state;
+}
+
 my_bool wsrep_thd_is_BF(void *thd_ptr, my_bool sync)
 { 
   my_bool status = FALSE;
@@ -453,7 +568,6 @@ my_bool wsrep_thd_is_local(void *thd_ptr, my_bool sync)
   return status;
 }
 
-extern "C"
 int wsrep_abort_thd(void *bf_thd_ptr, void *victim_thd_ptr, my_bool signal)
 {
   THD *victim_thd = (THD *) victim_thd_ptr;
@@ -461,7 +575,7 @@ int wsrep_abort_thd(void *bf_thd_ptr, void *victim_thd_ptr, my_bool signal)
   DBUG_ENTER("wsrep_abort_thd");
 
   if ( (WSREP(bf_thd) ||
-         ( (WSREP_ON || wsrep_OSU_method_options == WSREP_OSU_RSU) &&
+         ( (WSREP_ON || bf_thd->variables.wsrep_OSU_method == WSREP_OSU_RSU) &&
            bf_thd->wsrep_exec_mode == TOTAL_ORDER) )               &&
        victim_thd)
   {
@@ -477,7 +591,6 @@ int wsrep_abort_thd(void *bf_thd_ptr, void *victim_thd_ptr, my_bool signal)
   DBUG_RETURN(1);
 }
 
-extern "C"
 int wsrep_thd_in_locking_session(void *thd_ptr)
 {
   if (thd_ptr && ((THD *)thd_ptr)->in_lock_tables) {
@@ -486,3 +599,8 @@ int wsrep_thd_in_locking_session(void *thd_ptr)
   return 0;
 }
 
+bool wsrep_thd_has_explicit_locks(THD *thd)
+{
+  assert(thd);
+  return (thd->mdl_context.wsrep_has_explicit_locks());
+}
