@@ -1,4 +1,4 @@
-/* Copyright 2008 Codership Oy <http://www.codership.com>
+/* Copyright 2008-2015 Codership Oy <http://www.codership.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
 
 #include "wsrep_var.h"
 
@@ -22,13 +22,11 @@
 #include <sql_acl.h>
 #include "wsrep_priv.h"
 #include "wsrep_thd.h"
+#include "wsrep_xid.h"
 #include <my_dir.h>
 #include <cstdio>
 #include <cstdlib>
 
-#define WSREP_START_POSITION_ZERO "00000000-0000-0000-0000-000000000000:-1"
-#define WSREP_CLUSTER_NAME "my_wsrep_cluster"
-
 const  char* wsrep_provider         = 0;
 const  char* wsrep_provider_options = 0;
 const  char* wsrep_cluster_address  = 0;
@@ -37,8 +35,7 @@ const  char* wsrep_node_name        = 0;
 const  char* wsrep_node_address     = 0;
 const  char* wsrep_node_incoming_address = 0;
 const  char* wsrep_start_position   = 0;
-ulong  wsrep_reject_queries_options;
-ulong  wsrep_OSU_method_options;
+ulong   wsrep_reject_queries;
 
 int wsrep_init_vars()
 {
@@ -55,20 +52,74 @@ int wsrep_init_vars()
   return 0;
 }
 
+/* Function checks if the new value for wsrep_on is valid.
+Toggling of the value inside function or transaction is not allowed
+@return false if no error encountered with check else return true. */
+bool wsrep_on_check (sys_var *self, THD* thd, set_var* var)
+{
+  if (var->type == OPT_GLOBAL)
+    return true;
+
+  /* If in a stored function/trigger, it's too late to change wsrep_on. */
+  if (thd->in_sub_stmt)
+  {
+    WSREP_WARN("Cannot modify @@session.wsrep_on inside a stored function "
+               " or trigger");
+    my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+             var->save_result.ulonglong_value ? "ON" : "OFF");
+    return true;
+  }
+  /* Make the session variable 'wsrep_on' read-only inside a transaction. */
+  if (thd->in_active_multi_stmt_transaction())
+  {
+    WSREP_WARN("Cannot modify @@session.wsrep_on inside a transaction");
+    my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+             var->save_result.ulonglong_value ? "ON" : "OFF");
+    return true;
+  }
+
+  return false;
+}
+
 bool wsrep_on_update (sys_var *self, THD* thd, enum_var_type var_type)
 {
-  if (var_type == OPT_GLOBAL) {
-    // FIXME: this variable probably should be changed only per session
-    thd->variables.wsrep_on = global_system_variables.wsrep_on;
+  return false;
+}
+
+bool wsrep_causal_reads_update (sys_var *self, THD* thd, enum_var_type var_type)
+{
+  // global setting should not affect session setting.
+  // if (var_type == OPT_GLOBAL) {
+  //   thd->variables.wsrep_causal_reads = global_system_variables.wsrep_causal_reads;
+  // }
+  if (thd->variables.wsrep_causal_reads) {
+    thd->variables.wsrep_sync_wait |= WSREP_SYNC_WAIT_BEFORE_READ;
+  } else {
+    thd->variables.wsrep_sync_wait &= ~WSREP_SYNC_WAIT_BEFORE_READ;
+  }
+
+  // update global settings too.
+  if (global_system_variables.wsrep_causal_reads) {
+      global_system_variables.wsrep_sync_wait |= WSREP_SYNC_WAIT_BEFORE_READ;
+  } else {
+      global_system_variables.wsrep_sync_wait &= ~WSREP_SYNC_WAIT_BEFORE_READ;
   }
   return false;
 }
 
-void wsrep_causal_reads_update (sys_var *self, THD* thd, enum_var_type var_type)
+bool wsrep_sync_wait_update (sys_var* self, THD* thd, enum_var_type var_type)
 {
-  if (var_type == OPT_GLOBAL) {
-    thd->variables.wsrep_causal_reads = global_system_variables.wsrep_causal_reads;
-  }
+  // global setting should not affect session setting.
+  // if (var_type == OPT_GLOBAL) {
+  //   thd->variables.wsrep_sync_wait = global_system_variables.wsrep_sync_wait;
+  // }
+  thd->variables.wsrep_causal_reads = thd->variables.wsrep_sync_wait &
+          WSREP_SYNC_WAIT_BEFORE_READ;
+
+  // update global settings too
+  global_system_variables.wsrep_causal_reads = global_system_variables.wsrep_sync_wait &
+          WSREP_SYNC_WAIT_BEFORE_READ;
+  return false;
 }
 
 static int wsrep_start_position_verify (const char* start_str)
@@ -97,50 +148,53 @@ static int wsrep_start_position_verify (const char* start_str)
   return 1;
 }
 
+/* Function checks if the new value for start_position is valid.
+@return false if no error encountered with check else return true. */
 bool wsrep_start_position_check (sys_var *self, THD* thd, set_var* var)
 {
-  char   buff[FN_REFLEN];
-  String str(buff, sizeof(buff), system_charset_info), *res;
-  const char*   start_str = NULL;
+  char start_pos_buf[FN_REFLEN];
 
-  if (!(res = var->value->val_str(&str))) goto err;
+  if ((! var->save_result.string_value.str) ||
+      (var->save_result.string_value.length > (FN_REFLEN - 1))) // safety
+    goto err;
 
-  start_str = res->c_ptr();
+  memcpy(start_pos_buf, var->save_result.string_value.str,
+         var->save_result.string_value.length);
+  start_pos_buf[var->save_result.string_value.length]= 0;
 
-  if (!start_str) goto err;
-
-  if (!wsrep_start_position_verify(start_str)) return 0;
+  if (!wsrep_start_position_verify(start_pos_buf)) return false;
 
 err:
-
-  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str, 
-           start_str ? start_str : "NULL");
-  return 1;
+  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+           var->save_result.string_value.str ?
+           var->save_result.string_value.str : "NULL");
+  return true;
 }
 
-void wsrep_set_local_position (const char* value)
+static
+void wsrep_set_local_position(const char* const value, bool const sst)
 {
-  size_t value_len  = strlen (value);
-  size_t uuid_len   = wsrep_uuid_scan (value, value_len, &local_uuid);
+  size_t const value_len = strlen(value);
+  wsrep_uuid_t uuid;
+  size_t const uuid_len = wsrep_uuid_scan(value, value_len, &uuid);
+  wsrep_seqno_t const seqno = strtoll(value + uuid_len + 1, NULL, 10);
 
-  local_seqno = strtoll (value + uuid_len + 1, NULL, 10);
-
-  XID xid;
-  wsrep_xid_init(&xid, &local_uuid, local_seqno);
-  wsrep_set_SE_checkpoint(&xid);
-  WSREP_INFO ("wsrep_start_position var submitted: '%s'", wsrep_start_position);
+  if (sst) {
+    wsrep_sst_received (wsrep, uuid, seqno, NULL, 0);
+  } else {
+    // initialization
+    local_uuid = uuid;
+    local_seqno = seqno;
+  }
 }
 
 bool wsrep_start_position_update (sys_var *self, THD* thd, enum_var_type type)
 {
+  WSREP_INFO ("wsrep_start_position var submitted: '%s'",
+              wsrep_start_position);
   // since this value passed wsrep_start_position_check, don't check anything
   // here
-  wsrep_set_local_position (wsrep_start_position);
-
-  if (wsrep) {
-    wsrep_sst_received (wsrep, &local_uuid, local_seqno, NULL, 0);
-  }
-
+  wsrep_set_local_position (wsrep_start_position, true);
   return 0;
 }
 
@@ -153,7 +207,33 @@ void wsrep_start_position_init (const char* val)
     return;
   }
 
-  wsrep_set_local_position (val);
+  wsrep_set_local_position (val, false);
+}
+
+static int get_provider_option_value(const char* opts,
+                                     const char* opt_name,
+                                     ulong* opt_value)
+{
+  int ret= 1;
+  ulong opt_value_tmp;
+  char *opt_value_str, *s, *opts_copy= my_strdup(opts, MYF(MY_WME));
+
+  if ((opt_value_str= strstr(opts_copy, opt_name)) == NULL)
+    goto end;
+  opt_value_str= strtok_r(opt_value_str, "=", &s);
+  if (opt_value_str == NULL) goto end;
+  opt_value_str= strtok_r(NULL, ";", &s);
+  if (opt_value_str == NULL) goto end;
+
+  opt_value_tmp= strtoul(opt_value_str, NULL, 10);
+  if (errno == ERANGE) goto end;
+
+  *opt_value= opt_value_tmp;
+  ret= 0;
+
+end:
+  my_free(opts_copy);
+  return ret;
 }
 
 static bool refresh_provider_options()
@@ -163,9 +243,11 @@ static bool refresh_provider_options()
   char* opts= wsrep->options_get(wsrep);
   if (opts)
   {
-    if (wsrep_provider_options) my_free((void *)wsrep_provider_options);
-    wsrep_provider_options = (char*)my_memdup(opts, strlen(opts) + 1, 
-                                              MYF(MY_WME));
+    wsrep_provider_options_init(opts);
+    get_provider_option_value(wsrep_provider_options,
+                              (char*)"repl.max_ws_size",
+                              &wsrep_max_ws_size);
+    free(opts);
   }
   else
   {
@@ -190,7 +272,7 @@ static int wsrep_provider_verify (const char* provider_str)
     return 1;
 
   /* check that provider file exists */
-  bzero(&f_stat, sizeof(MY_STAT));
+  memset(&f_stat, 0, sizeof(MY_STAT));
   if (!my_stat(path, &f_stat, MYF(0)))
   {
     return 1;
@@ -198,25 +280,46 @@ static int wsrep_provider_verify (const char* provider_str)
   return 0;
 }
 
+/* Function checks if the new value for provider is valid.
+@return false if no error encountered with check else return true. */
 bool wsrep_provider_check (sys_var *self, THD* thd, set_var* var)
 {
-  char   buff[FN_REFLEN];
-  String str(buff, sizeof(buff), system_charset_info), *res;
-  const char*   provider_str = NULL;
+  char wsrep_provider_buf[FN_REFLEN];
 
-  if (!(res = var->value->val_str(&str))) goto err;
+  if ((! var->save_result.string_value.str) ||
+      (var->save_result.string_value.length > (FN_REFLEN - 1))) // safety
+    goto err;
 
-  provider_str = res->c_ptr();
+  /* Changing wsrep_provider in middle of function/trigger is not allowed. */
+  if (thd->in_sub_stmt)
+  {
+    WSREP_WARN("Cannot modify wsrep_provider inside a stored function "
+              " or trigger");
+    my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+             var->save_result.string_value.str);
+    return true;
+   }
 
-  if (!provider_str) goto err;
+  /* Changing wsrep_provider in middle of transaction is not allowed. */
+  if (thd->in_active_multi_stmt_transaction())
+  {
+    WSREP_WARN("Cannot modify wsrep_provider inside a transaction");
+    my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+             var->save_result.string_value.str);
+    return true;
+  }
 
-  if (!wsrep_provider_verify(provider_str)) return 0;
+  memcpy(wsrep_provider_buf, var->save_result.string_value.str,
+         var->save_result.string_value.length);
+  wsrep_provider_buf[var->save_result.string_value.length]= 0;
+
+  if (!wsrep_provider_verify(wsrep_provider_buf)) return false;
 
 err:
-
-  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str, 
-           provider_str ? provider_str : "NULL");
-  return 1;
+  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+           var->save_result.string_value.str ?
+           var->save_result.string_value.str : "NULL");
+  return true;
 }
 
 bool wsrep_provider_update (sys_var *self, THD* thd, enum_var_type type)
@@ -226,6 +329,14 @@ bool wsrep_provider_update (sys_var *self, THD* thd, enum_var_type type)
   bool wsrep_on_saved= thd->variables.wsrep_on;
   thd->variables.wsrep_on= false;
 
+  /* Ensure we free the stats that are allocated by galera-library.
+  wsrep part doesn't know how to free them shouldn't be attempting it
+  as it not allocated by wsrep part.
+  Before unloading cleanup any stale reference and stats is one of it.
+  Given that thd->variables.wsrep.on is turned-off it is safe to assume
+  that no new stats queries will be allowed during this transition time. */
+  wsrep_free_status(thd);
+
   WSREP_DEBUG("wsrep_provider_update: %s", wsrep_provider);
 
   /* stop replication is heavy operation, and includes closing all client 
@@ -237,7 +348,14 @@ bool wsrep_provider_update (sys_var *self, THD* thd, enum_var_type type)
   */
   mysql_mutex_unlock(&LOCK_global_system_variables);
   wsrep_stop_replication(thd);
+
+  /*
+    Unlock and lock LOCK_wsrep_slave_threads to maintain lock order & avoid
+    any potential deadlock.
+  */
+  mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
   mysql_mutex_lock(&LOCK_global_system_variables);
+  mysql_mutex_lock(&LOCK_wsrep_slave_threads);
 
   wsrep_deinit();
 
@@ -277,13 +395,19 @@ void wsrep_provider_init (const char* value)
   wsrep_provider = my_strdup(value, MYF(0));
 }
 
-bool wsrep_provider_options_check(sys_var *self, THD* thd, set_var* var)
+bool wsrep_provider_options_check (sys_var *self, THD* thd, set_var* var)
 {
   return 0;
 }
 
 bool wsrep_provider_options_update(sys_var *self, THD* thd, enum_var_type type)
 {
+  if (wsrep == NULL)
+  {
+    my_message(ER_WRONG_ARGUMENTS, "WSREP (galera) not started", MYF(0));
+    return true;
+  }
+
   wsrep_status_t ret= wsrep->options_set(wsrep, wsrep_provider_options);
   if (ret != WSREP_OK)
   {
@@ -294,59 +418,62 @@ bool wsrep_provider_options_update(sys_var *self, THD* thd, enum_var_type type)
   return refresh_provider_options();
 }
 
+void wsrep_provider_options_init(const char* value)
+{
+  if (wsrep_provider_options && wsrep_provider_options != value)
+    my_free((void *)wsrep_provider_options);
+  wsrep_provider_options = (value) ? my_strdup(value, MYF(0)) : NULL;
+}
+
 bool wsrep_reject_queries_update(sys_var *self, THD* thd, enum_var_type type)
 {
-    switch (wsrep_reject_queries_options) {
-        case WSREP_REJ_NONE:
-            wsrep_ready_set(TRUE); 
+    switch (wsrep_reject_queries) {
+        case WSREP_REJECT_NONE:
             WSREP_INFO("Allowing client queries due to manual setting");
             break;
-        case WSREP_REJ_ALL:
-            wsrep_ready_set(FALSE);
+        case WSREP_REJECT_ALL:
             WSREP_INFO("Rejecting client queries due to manual setting");
             break;
-        case WSREP_REJ_ALL_KILL:
-            wsrep_ready_set(FALSE);
+        case WSREP_REJECT_ALL_KILL:
             wsrep_close_client_connections(FALSE);
             WSREP_INFO("Rejecting client queries and killing connections due to manual setting");
             break;
         default:
-            WSREP_INFO("Unknown value");
+          WSREP_INFO("Unknown value for wsrep_reject_queries: %lu",
+                     wsrep_reject_queries);
             return true;
     }
     return false;
 }
 
-void wsrep_provider_options_init(const char* value)
-{
-  if (wsrep_provider_options && wsrep_provider_options != value) 
-    my_free((void *)wsrep_provider_options);
-  wsrep_provider_options = (value) ? my_strdup(value, MYF(0)) : NULL;
-}
-
 static int wsrep_cluster_address_verify (const char* cluster_address_str)
 {
   /* There is no predefined address format, it depends on provider. */
   return 0;
+
 }
 
+/* Function checks if the new value for cluster_address is valid.
+@return false if no error encountered with check else return true. */
 bool wsrep_cluster_address_check (sys_var *self, THD* thd, set_var* var)
 {
-  char   buff[FN_REFLEN];
-  String str(buff, sizeof(buff), system_charset_info), *res;
-  const char*   cluster_address_str = NULL;
+  char addr_buf[FN_REFLEN];
 
-  if (!(res = var->value->val_str(&str))) goto err;
+  if ((! var->save_result.string_value.str) ||
+      (var->save_result.string_value.length > (FN_REFLEN - 1))) // safety
+    goto err;
 
-  cluster_address_str = res->c_ptr();
+  memcpy(addr_buf, var->save_result.string_value.str,
+         var->save_result.string_value.length);
+  addr_buf[var->save_result.string_value.length]= 0;
 
-  if (!wsrep_cluster_address_verify(cluster_address_str)) return 0;
+  if (!wsrep_cluster_address_verify(addr_buf)) return false;
 
  err:
-
-  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str, 
-             cluster_address_str ? cluster_address_str : "NULL");
-  return 1    ;
+  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+           var->save_result.string_value.str ?
+           var->save_result.string_value.str : "NULL");
+  return true;
 }
 
 bool wsrep_cluster_address_update (sys_var *self, THD* thd, enum_var_type type)
@@ -363,7 +490,13 @@ bool wsrep_cluster_address_update (sys_var *self, THD* thd, enum_var_type type)
   */
   mysql_mutex_unlock(&LOCK_global_system_variables);
   wsrep_stop_replication(thd);
+  /*
+    Unlock and lock LOCK_wsrep_slave_threads to maintain lock order & avoid
+    any potential deadlock.
+  */
+  mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
   mysql_mutex_lock(&LOCK_global_system_variables);
+  mysql_mutex_lock(&LOCK_wsrep_slave_threads);
 
   if (wsrep_start_replication())
   {
@@ -386,25 +519,20 @@ void wsrep_cluster_address_init (const char* value)
   wsrep_cluster_address = (value) ? my_strdup(value, MYF(0)) : NULL;
 }
 
+/* Function checks if the new value for cluster_name is valid.
+@return false if no error encountered with check else return true. */
 bool wsrep_cluster_name_check (sys_var *self, THD* thd, set_var* var)
 {
-  char   buff[FN_REFLEN];
-  String str(buff, sizeof(buff), system_charset_info), *res;
-  const char* cluster_name_str = NULL;
-
-  if (!(res = var->value->val_str(&str))) goto err;
-
-  cluster_name_str = res->c_ptr();
-
-  if (!cluster_name_str || strlen(cluster_name_str) == 0) goto err;
-
-  return 0;
-
- err:
-
-  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str, 
-             cluster_name_str ? cluster_name_str : "NULL");
-  return 1;
+  if (!var->save_result.string_value.str ||
+      (var->save_result.string_value.length == 0) ||
+      (var->save_result.string_value.length > WSREP_CLUSTER_NAME_MAX_LEN))
+  {
+    my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+             (var->save_result.string_value.str ?
+              var->save_result.string_value.str : "NULL"));
+    return true;
+  }
+  return false;
 }
 
 bool wsrep_cluster_name_update (sys_var *self, THD* thd, enum_var_type type)
@@ -412,25 +540,19 @@ bool wsrep_cluster_name_update (sys_var *self, THD* thd, enum_var_type type)
   return 0;
 }
 
+/* Function checks if the new value for node_name is valid.
+@return false if no error encountered with check else return true. */
 bool wsrep_node_name_check (sys_var *self, THD* thd, set_var* var)
 {
-  char   buff[FN_REFLEN];
-  String str(buff, sizeof(buff), system_charset_info), *res;
-  const char* node_name_str = NULL;
-
-  if (!(res = var->value->val_str(&str))) goto err;
-
-  node_name_str = res->c_ptr();
-
-  if (!node_name_str || strlen(node_name_str) == 0) goto err;
-
-  return 0;
-
- err:
-
-  my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
-           node_name_str ? node_name_str : "NULL");
-  return 1;
+  // TODO: for now 'allow' 0-length string to be valid (default)
+  if (!var->save_result.string_value.str)
+  {
+    my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
+             (var->save_result.string_value.str ?
+              var->save_result.string_value.str : "NULL"));
+    return true;
+  }
+  return false;
 }
 
 bool wsrep_node_name_update (sys_var *self, THD* thd, enum_var_type type)
@@ -438,26 +560,29 @@ bool wsrep_node_name_update (sys_var *self, THD* thd, enum_var_type type)
   return 0;
 }
 
-// TODO: do something more elaborate, like checking connectivity
+/* Function checks if the new value for node_address is valid.
+TODO: do something more elaborate, like checking connectivity
+@return false if no error encountered with check else return true. */
 bool wsrep_node_address_check (sys_var *self, THD* thd, set_var* var)
 {
-  char   buff[FN_REFLEN];
-  String str(buff, sizeof(buff), system_charset_info), *res;
-  const char* node_address_str = NULL;
+  char addr_buf[FN_REFLEN];
 
-  if (!(res = var->value->val_str(&str))) goto err;
+  if ((! var->save_result.string_value.str) ||
+      (var->save_result.string_value.length > (FN_REFLEN - 1))) // safety
+    goto err;
 
-  node_address_str = res->c_ptr();
+  memcpy(addr_buf, var->save_result.string_value.str,
+         var->save_result.string_value.length);
+  addr_buf[var->save_result.string_value.length]= 0;
 
-  if (!node_address_str || strlen(node_address_str) == 0) goto err;
-
-  return 0;
-
- err:
+  // TODO: for now 'allow' 0-length string to be valid (default)
+  return false;
 
+err:
   my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), var->var->name.str,
-           node_address_str ? node_address_str : "NULL");
-  return 1;
+           var->save_result.string_value.str ?
+           var->save_result.string_value.str : "NULL");
+  return true;
 }
 
 bool wsrep_node_address_update (sys_var *self, THD* thd, enum_var_type type)
@@ -475,16 +600,9 @@ void wsrep_node_address_init (const char* value)
 
 bool wsrep_slave_threads_check (sys_var *self, THD* thd, set_var* var)
 {
-  if (wsrep_slave_count_change != 0) {
-      WSREP_ERROR("Still closing existing threads - %d", abs(wsrep_slave_count_change));
-     push_warning_printf(thd, MYSQL_ERROR::WARN_LEVEL_WARN,
-                        ER_UNKNOWN_ERROR,
-                        "Still closing existing threads - %d",
-                        abs(wsrep_slave_count_change));
-      return true;
-  }
   mysql_mutex_lock(&LOCK_wsrep_slave_threads);
-  wsrep_slave_count_change = var->value->val_int() - wsrep_slave_threads;
+  wsrep_slave_count_change += (var->save_result.ulonglong_value -
+                               wsrep_slave_threads);
   mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
 
   return false;
@@ -503,39 +621,33 @@ bool wsrep_slave_threads_update (sys_var *self, THD* thd, enum_var_type type)
 
 bool wsrep_desync_check (sys_var *self, THD* thd, set_var* var)
 {
-  bool new_wsrep_desync = var->save_result.ulonglong_value; 
+  bool new_wsrep_desync = var->save_result.ulonglong_value;
   if (wsrep_desync == new_wsrep_desync) {
     if (new_wsrep_desync) {
-      WSREP_DEBUG("wsrep_desync is already ON.");  
-      push_warning (thd, MYSQL_ERROR::WARN_LEVEL_WARN,
+      push_warning (thd, Sql_condition::WARN_LEVEL_WARN,
                    ER_WRONG_VALUE_FOR_VAR,
                    "'wsrep_desync' is already ON.");
-      return true;
     } else {
-      WSREP_DEBUG("wsrep_desync is already OFF.");  
-      push_warning (thd, MYSQL_ERROR::WARN_LEVEL_WARN,
+      push_warning (thd, Sql_condition::WARN_LEVEL_WARN,
                    ER_WRONG_VALUE_FOR_VAR,
                    "'wsrep_desync' is already OFF.");
-      return true;
     }
+    return false;
   }
-  return 0;
-}
-
-bool wsrep_desync_update (sys_var *self, THD* thd, enum_var_type type)
-{
   wsrep_status_t ret(WSREP_WARNING);
-  if (wsrep_desync) {
+  if (new_wsrep_desync) {
     ret = wsrep->desync (wsrep);
     if (ret != WSREP_OK) {
-      WSREP_WARN ("SET desync failed %d for %s", ret, thd->query());
+      WSREP_WARN ("SET desync failed %d for schema: %s, query: %s", ret,
+                  (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
       my_error (ER_CANNOT_USER, MYF(0), "'desync'", thd->query());
       return true;
     }
   } else {
     ret = wsrep->resync (wsrep);
     if (ret != WSREP_OK) {
-      WSREP_WARN ("SET resync failed %d for %s", ret, thd->query());
+      WSREP_WARN ("SET resync failed %d for schema: %s, query: %s", ret,
+                  (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
       my_error (ER_CANNOT_USER, MYF(0), "'resync'", thd->query());
       return true;
     }
@@ -543,6 +655,26 @@ bool wsrep_desync_update (sys_var *self, THD* thd, enum_var_type type)
   return false;
 }
 
+bool wsrep_desync_update (sys_var *self, THD* thd, enum_var_type type)
+{
+  return false;
+}
+
+bool wsrep_max_ws_size_update (sys_var *self, THD *thd, enum_var_type)
+{
+  char max_ws_size_opt[128];
+  my_snprintf(max_ws_size_opt, sizeof(max_ws_size_opt),
+              "repl.max_ws_size=%d", wsrep_max_ws_size);
+  wsrep_status_t ret= wsrep->options_set(wsrep, max_ws_size_opt);
+  if (ret != WSREP_OK)
+  {
+    WSREP_ERROR("Set options returned %d", ret);
+    refresh_provider_options();
+    return true;
+  }
+  return refresh_provider_options();
+}
+
 /*
  * Status variables stuff below
  */
@@ -579,6 +711,12 @@ static void export_wsrep_status_to_mysql(THD* thd)
 {
   int wsrep_status_len, i;
 
+  /* Avoid freeing the stats immediately on completion of the call
+  instead free them on next invocation as the reference to status
+  is feeded in MySQL show status array. Freeing them on completion
+  will make these references invalid. */
+  wsrep_free_status(thd);
+
   thd->wsrep_status_vars = wsrep->stats_get(wsrep);
 
   if (!thd->wsrep_status_vars) {
@@ -587,7 +725,9 @@ static void export_wsrep_status_to_mysql(THD* thd)
 
   for (wsrep_status_len = 0;
        thd->wsrep_status_vars[wsrep_status_len].name != NULL;
-       wsrep_status_len++);
+       wsrep_status_len++) {
+      /* */
+  }
 
 #if DYNAMIC
   if (wsrep_status_len != mysql_status_len) {
